#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-06-10
# @Filename: mwm_yso.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    MIPSGAL,
    AllWise,
    Catalog,
    CatalogToAllWise,
    CatalogToGaia_DR3,
    CatalogToTIC_v8,
    CatalogToTwoMassPSC,
    Gaia_DR2,
    Gaia_DR3,
    Sagitta_EDR3,
    TIC_v8,
    TwoMassPSC,
    YSO_Clustering,
    Zari18pms,
)

from target_selection.cartons import BaseCarton
from target_selection.exceptions import TargetSelectionError


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py


class MWM_YSO_Disk_APOGEE_Carton(BaseCarton):
    """YSOs - Disk APOGEE (IR excess).

    Shorthand name: mwm_yso_disk_apogee

    old class name: MWM_YSO_S1_Carton
    old shorthand name: mwm_yso_s1

    Simplified Description of selection criteria:
    selection of YSOs based on IR excess,
    with WISE colors W1-W2>0.25, W2-W3>0.5, W3-W4>1.5,
    closer than parallax>0.3, and brighter than H<13
    (should have ~21.5K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Gaia, 2mass, allwise
    Additional cross-matching needed:
    Note: Using the Gaia xmatch somehow misses half the sources.
    Selection was done on the allwise catalog that
    had 2mass photometry,
    and then the resulting selection was crossmatched against against
    Gaia with 1" search radius.
    Return columns: Gaia id, 2mass id, allwise id, G, BP, RP,
    J, H, K, W1, W2, W3, W4,parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: h_m<13 and w1mpro-w2mpro>0.25 and
    w2mpro-w3mpro>0.5 and w3mpro-w4mpro>1.5 and parallax>0.3

    """

    name = "mwm_yso_disk_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                AllWise.designation.alias("allwise_designation"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag.alias("gaia_dr3_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR3.parallax,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(CatalogToGaia_DR3)
            .join(CatalogToAllWise, on=(CatalogToGaia_DR3.catalogid == CatalogToAllWise.catalogid))
            .join(AllWise, on=(CatalogToAllWise.target_id == AllWise.cntr))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.best >> True,
                CatalogToAllWise.best >> True,
                TwoMassPSC.h_m < 13,
                (AllWise.w1mpro - AllWise.w2mpro) > 0.25,
                (AllWise.w2mpro - AllWise.w3mpro) > 0.50,
                (AllWise.w3mpro - AllWise.w4mpro) > 1.50,
                Gaia_DR3.parallax > 0.3,
            )
        )

        # There can be cases in which the same catalogid has multiple entries
        # in a catalog_to_x table since the same physical object
        # may match with multiple catalogids.
        # Hence, we have the below code in the above query.
        #                CatalogToGaia_DR3.best >> True,
        #                CatalogToTwoMassPSC.best >> True,
        #                CatalogToAllWise.best >> True,

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_YSO_Disk_APOGEE_Single_Carton(MWM_YSO_Disk_APOGEE_Carton):
    name = "mwm_yso_disk_apogee_single"
    cadence = "bright_1x1"
    priority = 2706


class MWM_YSO_Disk_BOSS_Carton(BaseCarton):
    """YSOs - Disk BOSS (IR excess).

    Shorthand name: mwm_yso_disk_boss

    old class name: MWM_YSO_S1_Carton
    old shorthand name: mwm_yso_s1

    Simplified Description of selection criteria:
    selection of YSOs based on IR excess,
    with WISE colors W1-W2>0.25, W2-W3>0.5, W3-W4>1.5,
    closer than parallax>0.3, and brighter than H<13
    (should have ~21.5K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Gaia, 2mass, allwise
    Additional cross-matching needed:
    Note: Using the Gaia xmatch somehow misses half the sources.
    Selection was done on the allwise catalog that
    had 2mass photometry,
    and then the resulting selection was crossmatched against against
    Gaia with 1" search radius.
    Return columns: Gaia id, 2mass id, allwise id, G, BP, RP,
    J, H, K, W1, W2, W3, W4,parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation: phot_rp_mean_mag<15.5 and w1mpro-w2mpro>0.25 and
    w2mpro-w3mpro>0.5 and w3mpro-w4mpro>1.5 and parallax>0.3

    Comments: Split from mwm_yso_s1 to request BOSS observations,
    same color selection but assigning cadence and faint limit for carton based
    on RP instead of H
    """

    name = "mwm_yso_disk_boss"
    category = "science"
    instrument = None  # instrument is set in post_process()
    cadence = None  # cadence is set in post_process()
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                AllWise.designation.alias("allwise_designation"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag.alias("gaia_dr3_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR3.parallax,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(CatalogToGaia_DR3)
            .join(CatalogToAllWise, on=(CatalogToGaia_DR3.catalogid == CatalogToAllWise.catalogid))
            .join(AllWise, on=(CatalogToAllWise.target_id == AllWise.cntr))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.best >> True,
                CatalogToAllWise.best >> True,
                Gaia_DR3.phot_rp_mean_mag < 15.5,
                (AllWise.w1mpro - AllWise.w2mpro) > 0.25,
                (AllWise.w2mpro - AllWise.w3mpro) > 0.50,
                (AllWise.w3mpro - AllWise.w4mpro) > 1.50,
                Gaia_DR3.parallax > 0.3,
            )
        )

        # Gaia_DR3 pweewee model class corresponds to
        # table catalogdb.gaia_dr3_source.
        #

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr3_rp from " + " sandbox.temp_mwm_yso_disk_boss ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if current_rp < 14.76:
                current_instrument = "BOSS"
                current_cadence = "bright_3x1"
            elif current_rp < 15.075:
                current_instrument = "BOSS"
                current_cadence = "bright_4x1"
            elif current_rp < 15.29:
                current_instrument = "BOSS"
                current_cadence = "bright_5x1"
            elif current_rp < 15.5:
                current_instrument = "BOSS"
                current_cadence = "bright_6x1"
            else:
                # All cases should be covered above so we should not get here.
                current_instrument = None
                current_cadence = None
                raise TargetSelectionError(
                    "error in mwm_yso_disk_boss "
                    + "post_process(): "
                    + "instrument = None, cadence= None"
                )

            if current_instrument is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_disk_boss "
                    + " set instrument = '"
                    + current_instrument
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_disk_boss "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )


class MWM_YSO_Disk_BOSS_Single_Carton(MWM_YSO_Disk_BOSS_Carton):
    name = "mwm_yso_disk_boss_single"
    cadence = "bright_1x1"
    priority = 2706
    instrument = "BOSS"

    def post_process(self, model):
        pass


# The below class is from commit of Feb 2, 2022.
# The only changes are
# (1) can_offset = True
# (2) addition of the
# last condition in the WHERE clause and
# (3) addition of TwoMassPSC.ra and TwoMassPSC.decl
# in the SELECT clause.
class MWM_YSO_Embedded_APOGEE_Carton(BaseCarton):
    """YSOs - Embedded APOGEE (optically invisible).

    Shorthand name: mwm_yso_embedded_apogee

    old class name: MWM_YSO_S2_Carton
    old shorthand name: mwm_yso_s2

    Wiki page: v0.5
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass+allwise, gaia
    (allow sources that lack gaia xmatch)
    Additional cross-matching needed:
    Note: Using the Gaia xmatch somehow misses half the sources.
    Selection was done on the allwise catalog
    that had 2mass photometry,
    and then the resulting selection was crossmatched
    against against Gaia with 1" search radius.
    Return columns: Gaia id, 2mass id, allwise id, G, BP, RP,
    J, H, K, W1, W2, W3, W4
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):

    Pseudo SQL (optional): v1.0
    Implementation: h_m<13 and
    (phot_g_mean_mag>18.5 or phot_g_mean_mag is null)
    and j_m-h_m>1
    and h_m-ks_m>0.5 << v0.5 and v1.0 condition on H-K
    and w1mpro-w2mpro>0.5
    and w2mpro-w3mpro>1
    and w3mpro-w4mpro>1.5
    and w3mpro-w4mpro>(w1mpro-w2mpro)*0.8+1.1
    and (H-K)>0.65*(J-H)-0.25 <<< This is additional condition on H-K
                                  compared to v0.5
    """

    name = "mwm_yso_embedded_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            AllWise.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                TwoMassPSC.ra.alias("twomass_psc_ra"),
                TwoMassPSC.decl.alias("twomass_psc_decl"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                AllWise.designation.alias("allwise_designation"),
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag.alias("gaia_dr2_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR2.parallax,
            )
            .join(TIC_v8, on=(TIC_v8.allwise == AllWise.designation))
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .switch(TIC_v8)
            .join(Gaia_DR2, peewee.JOIN.LEFT_OUTER, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .switch(TIC_v8)
            .join(CatalogToTIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                TwoMassPSC.h_m < 13,
                (Gaia_DR2.phot_g_mean_mag > 18.5) | (Gaia_DR2.phot_g_mean_mag >> None),
                (AllWise.j_m_2mass - AllWise.h_m_2mass) > 1.0,
                (AllWise.h_m_2mass - AllWise.k_m_2mass) > 0.5,
                (AllWise.w1mpro - AllWise.w2mpro) > 0.50,
                (AllWise.w2mpro - AllWise.w3mpro) > 1.00,
                (AllWise.w3mpro - AllWise.w4mpro) > 1.50,
                (AllWise.w3mpro - AllWise.w4mpro) > (AllWise.w1mpro - AllWise.w2mpro) * 0.8 + 1.1,
                (
                    (AllWise.h_m_2mass - AllWise.k_m_2mass)  # added for v1.0
                    > (0.65 * (AllWise.j_m_2mass - AllWise.h_m_2mass) - 0.25)
                ),
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class MWM_YSO_Embedded_APOGEE_Single_Carton(MWM_YSO_Embedded_APOGEE_Carton):
    name = "mwm_yso_embedded_apogee_single"
    cadence = "bright_1x1"
    priority = 2706


class MWM_YSO_Nebula_APOGEE_Carton(BaseCarton):
    """YSOs - Nebula APOGEE(optically invisible, WISE saturated).

    Shorthand name: mwm_yso_nebula_apogee

    old class name: MWM_YSO_S2_5_Carton
    old shorthand name: mwm_yso_s2_5

    Simplified Description of selection criteria:
    selection of YSOs, brighter than H<15,
    saturated (blank) W4 with W2-W3>4,
    or saturated W3 and W2, with J-H>1.1.
    Some contaminants from scanning are
    filtered on the plane of the sky:
    all the targets should be within 5 deg of the plane+
    few sources that can be located
    further south of the plane if l>180
    (should have ~1.2K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, allwise
    Additional cross-matching needed:
    Return columns: 2mass id, allwise id, J, H, K, W1, W2, W3, W4
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: h_m<13 and
    (w2mpro-w3mpro>4 and w4mpro is null) or
    (w3mpro is null and w4mpro is null and j_m-h_m>1.1)
    and (b>-5 or l>180) and b<-5

    """

    name = "mwm_yso_nebula_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    # Above implementation has below clause
    # and (b>-5 or l>180) and b<-5
    # Replace (b>-5 or l>180) and b<-5 as below based on the text.
    # In words:
    # all the targets should be within 5 deg of the plane+
    # few sources that can be
    # located further south of the plane if l>180
    # Hence:
    # ((b>-5) and (b<5)) or ((b<-5) and (l > 180))
    #  l, b in Gaia_DR2 are gallong and gallat in TIC_v8.
    # We are using the values from Gaia since
    # TIC propagates the coordinates back to epoch 2000.0
    # (b>-5 or l>180) and b<-5
    # S2_5 query below has the same part before where() as S2 query.

    def build_query(self, version_id, query_region=None):
        query = (
            AllWise.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                TwoMassPSC.ra.alias("twomass_psc_ra"),
                TwoMassPSC.decl.alias("twomass_psc_decl"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                AllWise.designation.alias("allwise_designation"),
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag.alias("gaia_dr2_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR2.parallax,
            )
            .join(TIC_v8, on=(TIC_v8.allwise == AllWise.designation))
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .switch(TIC_v8)
            .join(Gaia_DR2, peewee.JOIN.LEFT_OUTER, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .switch(TIC_v8)
            .join(CatalogToTIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                TwoMassPSC.h_m < 13,
                (((AllWise.w2mpro - AllWise.w3mpro) > 4) & (AllWise.w4mpro >> None))
                | (
                    (AllWise.w3mpro >> None)
                    & (AllWise.w4mpro >> None)
                    & ((AllWise.j_m_2mass - AllWise.h_m_2mass) > 1.1)
                ),
                ((Gaia_DR2.b > -5) & (Gaia_DR2.b < 5))
                | ((Gaia_DR2.b < -5) & (Gaia_DR2.l > 180))
                | ((Gaia_DR2.b >> None) & (Gaia_DR2.l >> None)),
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class MWM_YSO_Nebula_APOGEE_Single_Carton(MWM_YSO_Nebula_APOGEE_Carton):
    name = "mwm_yso_nebula_apogee_single"
    cadence = "bright_1x1"
    priority = 2706


class MWM_YSO_Variable_APOGEE_Carton(BaseCarton):
    """YSOs - Variable APOGEE (pre-main sequence optical variables).

    Shorthand name: mwm_yso_variable_apogee

    old class name: MWM_YSO_S3_Carton
    old shorthand name: mwm_yso_s3

    Simplified Description of selection criteria:
    selection of YSOs brighter than H<13, closer than parallax>0.3.
    Filter on the position of the HR diagram to
    select cool pre-main sequence stars,
    with BP-RP>13, (BP-RP)*2.5+2.5>M_G, (BP-RP)*2.5-1<M_G,
    requiring variability in g,bp,rp>0.02
    (with var_x defined as
    sqrt(phot_x_n_obs)/phot_x_mean_flux_over_error),
    have relations in variability of
    var_g<var_bp<var_g^0.75, 0.75*var_g<var_rp<var_g^0.95,
    and log10(var_bp)*5+11<M_BP, in which M_x is the absolute mag
    (should have ~52.7K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, gaia
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation:
    phot_g_mean_mag < 18.5 and h_m <13 and parallax >0.3 and
    bp_rp*2.5+2.5 > phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp*2.5-1 < phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error*0.75 and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.75) and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.95) and
    log10(sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error)*5+11<
    phot_bp_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp>1.3 and sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error>0.02
    and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>0.02 and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>0.02

    """

    name = "mwm_yso_variable_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                AllWise.designation.alias("allwise_designation"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag.alias("gaia_dr3_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR3.parallax,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(CatalogToGaia_DR3)
            .join(CatalogToAllWise, on=(CatalogToGaia_DR3.catalogid == CatalogToAllWise.catalogid))
            .join(AllWise, on=(CatalogToAllWise.target_id == AllWise.cntr))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.best >> True,
                CatalogToAllWise.best >> True,
                Gaia_DR3.phot_g_mean_mag < 18.5,
                TwoMassPSC.h_m < 13,
                Gaia_DR3.parallax > 0.3,
                Gaia_DR3.bp_rp * 2.5 + 2.5
                > Gaia_DR3.phot_g_mean_mag - 5 * (peewee.fn.log(1000 / Gaia_DR3.parallax) - 1),
                Gaia_DR3.bp_rp * 2.5 - 1
                < Gaia_DR3.phot_g_mean_mag - 5 * (peewee.fn.log(1000 / Gaia_DR3.parallax) - 1),
                peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                > peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error,
                peewee.fn.sqrt(Gaia_DR3.phot_rp_n_obs) / Gaia_DR3.phot_rp_mean_flux_over_error
                > peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs)
                / Gaia_DR3.phot_g_mean_flux_over_error
                * 0.75,
                peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                < peewee.fn.power(
                    peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error,
                    0.75,
                ),
                peewee.fn.sqrt(Gaia_DR3.phot_rp_n_obs) / Gaia_DR3.phot_rp_mean_flux_over_error
                < peewee.fn.power(
                    peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error,
                    0.95,
                ),
                peewee.fn.log(
                    peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                )
                * 5
                + 11
                < Gaia_DR3.phot_bp_mean_mag - 5 * (peewee.fn.log(1000 / Gaia_DR3.parallax) - 1),
                Gaia_DR3.bp_rp > 1.3,
                peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error
                > 0.02,
                peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                > 0.02,
                peewee.fn.sqrt(Gaia_DR3.phot_rp_n_obs) / Gaia_DR3.phot_rp_mean_flux_over_error
                > 0.02,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class MWM_YSO_Variable_APOGEE_Single_Carton(MWM_YSO_Variable_APOGEE_Carton):
    name = "mwm_yso_variable_apogee_single"
    cadence = "bright_1x1"
    priority = 2706


class MWM_YSO_Variable_BOSS_Carton(BaseCarton):
    """YSOs - Variable BOSS (pre-main sequence optical variables).

    Shorthand name: mwm_yso_variable_boss

    old class name: MWM_YSO_S3_Carton
    old shorthand name: mwm_yso_s3

    Simplified Description of selection criteria:
    selection of YSOs brighter than H<13, closer than parallax>0.3.
    Filter on the position of the HR diagram to
    select cool pre-main sequence stars,
    with BP-RP>13, (BP-RP)*2.5+2.5>M_G, (BP-RP)*2.5-1<M_G,
    requiring variability in g,bp,rp>0.02
    (with var_x defined as
    sqrt(phot_x_n_obs)/phot_x_mean_flux_over_error),
    have relations in variability of
    var_g<var_bp<var_g^0.75, 0.75*var_g<var_rp<var_g^0.95,
    and log10(var_bp)*5+11<M_BP, in which M_x is the absolute mag
    (should have ~52.7K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: 2mass, gaia
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation:
    phot_rp_mean_mag<15.5 and phot_g_mean_mag < 18.5 and h_m <13 and parallax >0.3 and
    bp_rp*2.5+2.5 > phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp*2.5-1 < phot_g_mean_mag-5*(log10(1000/parallax)-1) and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>
    sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error*0.75 and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.75) and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error<
    power(sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error,0.95) and
    log10(sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error)*5+11<
    phot_bp_mean_mag-5*(log10(1000/parallax)-1) and
    bp_rp>1.3 and sqrt(phot_g_n_obs)/phot_g_mean_flux_over_error>0.02
    and
    sqrt(phot_bp_n_obs)/phot_bp_mean_flux_over_error>0.02 and
    sqrt(phot_rp_n_obs)/phot_rp_mean_flux_over_error>0.02

    Comments: Split from mwm_yso_s3 to request BOSS observations,
    RP magnitude check added to the previous selection
    """

    name = "mwm_yso_variable_boss"
    category = "science"
    instrument = None  # instrument is set in post_process()
    cadence = None  # cadence is set in post_process()
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                AllWise.designation.alias("allwise_designation"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag.alias("gaia_dr3_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR3.parallax,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(CatalogToGaia_DR3)
            .join(CatalogToAllWise, on=(CatalogToGaia_DR3.catalogid == CatalogToAllWise.catalogid))
            .join(AllWise, on=(CatalogToAllWise.target_id == AllWise.cntr))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.best >> True,
                CatalogToAllWise.best >> True,
                Gaia_DR3.phot_rp_mean_mag < 15.5,
                Gaia_DR3.phot_g_mean_mag < 18.5,
                TwoMassPSC.h_m < 13,
                Gaia_DR3.parallax > 0.3,
                Gaia_DR3.bp_rp * 2.5 + 2.5
                > Gaia_DR3.phot_g_mean_mag - 5 * (peewee.fn.log(1000 / Gaia_DR3.parallax) - 1),
                Gaia_DR3.bp_rp * 2.5 - 1
                < Gaia_DR3.phot_g_mean_mag - 5 * (peewee.fn.log(1000 / Gaia_DR3.parallax) - 1),
                peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                > peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error,
                peewee.fn.sqrt(Gaia_DR3.phot_rp_n_obs) / Gaia_DR3.phot_rp_mean_flux_over_error
                > peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs)
                / Gaia_DR3.phot_g_mean_flux_over_error
                * 0.75,
                peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                < peewee.fn.power(
                    peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error,
                    0.75,
                ),
                peewee.fn.sqrt(Gaia_DR3.phot_rp_n_obs) / Gaia_DR3.phot_rp_mean_flux_over_error
                < peewee.fn.power(
                    peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error,
                    0.95,
                ),
                peewee.fn.log(
                    peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                )
                * 5
                + 11
                < Gaia_DR3.phot_bp_mean_mag - 5 * (peewee.fn.log(1000 / Gaia_DR3.parallax) - 1),
                Gaia_DR3.bp_rp > 1.3,
                peewee.fn.sqrt(Gaia_DR3.phot_g_n_obs) / Gaia_DR3.phot_g_mean_flux_over_error
                > 0.02,
                peewee.fn.sqrt(Gaia_DR3.phot_bp_n_obs) / Gaia_DR3.phot_bp_mean_flux_over_error
                > 0.02,
                peewee.fn.sqrt(Gaia_DR3.phot_rp_n_obs) / Gaia_DR3.phot_rp_mean_flux_over_error
                > 0.02,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr3_rp from " + " sandbox.temp_mwm_yso_variable_boss ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if current_rp < 14.76:
                current_instrument = "BOSS"
                current_cadence = "bright_3x1"
            elif current_rp < 15.075:
                current_instrument = "BOSS"
                current_cadence = "bright_4x1"
            elif current_rp < 15.29:
                current_instrument = "BOSS"
                current_cadence = "bright_5x1"
            elif current_rp < 15.5:
                current_instrument = "BOSS"
                current_cadence = "bright_6x1"
            else:
                # All cases should be covered above so we should not get here.
                current_instrument = None
                current_cadence = None
                raise TargetSelectionError(
                    "error in mwm_yso_variable_boss "
                    + "post_process(): "
                    + "instrument = None, cadence= None"
                )

            if current_instrument is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_variable_boss "
                    + " set instrument = '"
                    + current_instrument
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_variable_boss "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )


class MWM_YSO_Variable_BOSS_Single_Carton(MWM_YSO_Variable_BOSS_Carton):
    name = "mwm_yso_variable_boss_single"
    cadence = "bright_1x1"
    priority = 2706
    instrument = "BOSS"

    def post_process(self, model):
        pass


class MWM_YSO_CMZ_APOGEE_Carton(BaseCarton):
    """YSOs - Central Molecular Zone APOGEE.

    Shorthand name: mwm_yso_cmz_apogee

    old class name: MWM_YSO_CMZ_Carton
    old shorthand name: mwm_yso_cmz

    Simplified Description of selection criteria:
    selection of sources in the central molecular zone
    based on spitzer fluxes from mipsgal.
     brighter than H<13, have color 8.0-24>2.5, and
     have parallax<0.2 or lack a Gaia xmatch.
    (should have ~3.2K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: mipsgal
    Additional cross-matching needed: the table has xmatch included
    Return columns:
    mipsgal id, 2mass id, j, h, k, 3.6, 4.8, 8.0, 24 mag
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    'apogee_bright_3x1'
    Pseudo SQL (optional):
    Implementation: Hmag<13 and _8_0_-_24_>2.5 and
    (parallax<0.2 or parallax is null)

    DONE (email of 1/6/20232 with jsg) check left outer join
    TODO Below comments are for v0.5. modify for v1.0.

    For CMZ, the raw sql query would be:
    select ct.catalogid from mipsgal m
    join twomass_psc t on twomass_name = designation
    join tic_v8 tic on tic.twomass_psc = t.designation
    left outer join gaia_dr2_source g on g.source_id = tic.gaia_int
    join catalog_to_tic_v8 ct on ct.target_id = tic.id
    where m.hmag < 13 and
    (m.mag_8_0 - m.mag_24) > 2.5 and
    (g.parallax < 0.2 or g.parallax is null)
    and ct.version_id = 13 and ct.best is true;

    Note you only need one left outer join between TIC and Gaia
    (all MIPSGAL targets have a counterpart in 2MASS,
    and all 2MASS have an entry in TIC,
    but not all the TIC entries have a Gaia counterpart).

    Comments: Formerly mwm_yso_cmz, removed check on the position on the sky:

    Removed below condition.
    l is glon (galactic longitude)
    b is glat (galactic latitude)
    All four statements below are equivalent.
    (l> 358 or l< 2) and
    b between -1 and 1

    (m.glon > 358 or m.glon < 2) and
    (m.glat > -1 and m.glat < 1) and

    Sources are within 2 degrees in l and
     1 degree in b from the galactic center,

    (MIPSGAL.glon > 358) | (MIPSGAL.glon < 2),
    (MIPSGAL.glat > -1) & (MIPSGAL.glat < 1),

    """

    name = "mwm_yso_cmz_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    # mipsgal is a subset of 2MASS
    # mipsgal can be joined to twomass_psc via
    # mipsgal.twomass_name = TwoMassPSC.designation.
    #
    # mipsgal is a subset of 2MASS
    #
    # 2MASS is not a subset of Gaia_DR3
    # Gaia_DR3 is not a subset of 2MASS
    #
    # table catalogdb.mipsgal
    # Foreign-key constraints:
    #    "twomass_name_fk" FOREIGN KEY (twomass_name)
    # REFERENCES twomass_psc(designation)
    #
    # Since the below condition is true,
    # we do not need a between to Catalog and CatalogToTwoMassPSC
    #
    # Catalog.catalogid == CatalogToTwoMassPSC.catalogid
    #
    # We can remove the join with Catalog in all the cartons
    # since catalogid is completely unique (even across different version_id)
    # so the join with Catalog doesn't give us anything extra and it's a costly join.

    def build_query(self, version_id, query_region=None):
        # after the first left outer join, the later joins must
        # also be left outer joins
        query = (
            CatalogToTwoMassPSC.select(
                CatalogToTwoMassPSC.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                MIPSGAL.mag_3_6,
                MIPSGAL.mag_4_5,
                MIPSGAL.mag_5_8,
                MIPSGAL.mag_8_0,
                MIPSGAL.mag_24,
                MIPSGAL.hmag,
                Gaia_DR3.parallax,
                MIPSGAL.glon,
                MIPSGAL.glat,
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .join(MIPSGAL, on=(MIPSGAL.twomass_name == TwoMassPSC.designation))
            .switch(CatalogToTwoMassPSC)
            .join(
                CatalogToGaia_DR3,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToTwoMassPSC.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(
                Gaia_DR3,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id),
            )
            .distinct(CatalogToTwoMassPSC.catalogid)
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,  # See below for CatalogToGaia_DR3.best
                MIPSGAL.hmag < 13,
                (MIPSGAL.mag_8_0 - MIPSGAL.mag_24) > 2.5,
                ((Gaia_DR3.parallax < 0.2) & (CatalogToGaia_DR3.best >> True))
                | (Gaia_DR3.parallax >> None),
            )
        )
        # above condition (Gaia_DR3.parallax >> None) ensures that
        # that we get the rows from the left outer join

        if query_region:
            query = query.join_from(CatalogToTwoMassPSC, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_YSO_CMZ_APOGEE_Single_Carton(MWM_YSO_CMZ_APOGEE_Carton):
    name = "mwm_yso_cmz_apogee_single"
    cadence = "bright_1x1"
    priority = 2706


class MWM_YSO_Cluster_APOGEE_Carton(BaseCarton):
    """YSOs - Cluster APOGEE Catalog
    Shorthand name: mwm_yso_cluster_apogee

    old class name: MWM_YSO_Cluster_Carton
    old shorthand name: mwm_yso_cluster

    Simplified Description of selection criteria:
    Selecting the clustered sources from
    the catalog of clustered structures,
    with age<7.5 dex and brighter than H<13 mag.
    (should have ~45.5K sources)
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Kounkel+20 clustered catalog
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    Pseudo SQL (optional):
    Implementation: age<7.5 and h<13

    """

    name = "mwm_yso_cluster_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    # yso_clustering is a subset of gaia and
    # can be joined to gaia_dr2_source via source_id.
    #
    # table catalogdb.yso_clustering
    # Foreign-key constraints:
    #    "yso_clustering_source_id_fkey" FOREIGN KEY (source_id)
    # REFERENCES gaia_dr2_source(source_id)

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToTIC_v8.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                YSO_Clustering.twomass,
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag.alias("gaia_dr2_rp"),
                YSO_Clustering.j,
                YSO_Clustering.h,
                YSO_Clustering.k,
                Gaia_DR2.parallax,
            )
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .join(YSO_Clustering, on=(Gaia_DR2.source_id == YSO_Clustering.source_id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                YSO_Clustering.h < 13,
                YSO_Clustering.age < 7.5,
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_YSO_Cluster_APOGEE_Single_Carton(MWM_YSO_Cluster_APOGEE_Carton):
    name = "mwm_yso_cluster_apogee_single"
    cadence = "bright_1x1"
    priority = 2706


class MWM_YSO_Cluster_BOSS_Carton(BaseCarton):
    """YSOs - Cluster BOSS Catalog
    Shorthand name: mwm_yso_cluster_boss

    old class name: MWM_YSO_Cluster_Carton
    old shorthand name: mwm_yso_cluster

    Simplified Description of selection criteria:
    Selecting the clustered sources from
    the catalog of clustered structures,
    with age<7.5 dex and brighter than rp<15.5 mag.

    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    Additional source catalogs needed: Kounkel+20 clustered catalog
    Additional cross-matching needed:
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options,
    even though no single target will receive more than one):
    cadence options for these targets:
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Pseudo SQL (optional):
    Implementation: age<7.5 and rp<15.5

    Comments: Split from Cluster to request BOSS observations,
    assigning cadence and faint limit for carton based on RP instead of H
    """

    name = "mwm_yso_cluster_boss"
    category = "science"
    instrument = None  # instrument is set in post_process()
    cadence = None  # cadence is set in post_process()
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2705
    can_offset = True

    # yso_clustering is a subset of gaia and
    # can be joined to gaia_dr2_source via source_id.
    #
    # table catalogdb.yso_clustering
    # Foreign-key constraints:
    #    "yso_clustering_source_id_fkey" FOREIGN KEY (source_id)
    # REFERENCES gaia_dr2_source(source_id)

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToTIC_v8.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                YSO_Clustering.twomass,
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag.alias("gaia_dr2_rp"),
                YSO_Clustering.j,
                YSO_Clustering.h,
                YSO_Clustering.k,
                Gaia_DR2.parallax,
            )
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .join(YSO_Clustering, on=(Gaia_DR2.source_id == YSO_Clustering.source_id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                Gaia_DR2.phot_rp_mean_mag < 15.5,
                YSO_Clustering.age < 7.5,
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " + " sandbox.temp_mwm_yso_cluster_boss ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if current_rp < 14.76:
                current_instrument = "BOSS"
                current_cadence = "bright_3x1"
            elif current_rp < 15.075:
                current_instrument = "BOSS"
                current_cadence = "bright_4x1"
            elif current_rp < 15.29:
                current_instrument = "BOSS"
                current_cadence = "bright_5x1"
            elif current_rp < 15.5:
                current_instrument = "BOSS"
                current_cadence = "bright_6x1"
            else:
                # All cases should be covered above so we should not get here.
                current_instrument = None
                current_cadence = None
                raise TargetSelectionError(
                    "error in mwm_yso_cluster_boss "
                    + "post_process(): "
                    + "instrument = None, cadence= None"
                )

            if current_instrument is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_cluster_boss "
                    + " set instrument = '"
                    + current_instrument
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_cluster_boss "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )


class MWM_YSO_Cluster_BOSS_Single_Carton(MWM_YSO_Cluster_BOSS_Carton):
    name = "mwm_yso_cluster_boss_single"
    cadence = "bright_1x1"
    priority = 2706
    instrument = "BOSS"

    def post_process(self, model):
        pass


class MWM_YSO_PMS_APOGEE_Sagitta_EDR3_Carton(BaseCarton):
    """
    YSOs - Pre-main sequence, APOGEE
    Shorthand name: mwm_yso_pms_apogee_sagitta_edr3
    Comments: For v1.0, we use sagitta_edr3 instead of sagitta.
    Simplified Description of selection criteria:
    Selecting the clustered sources from the catalog of vetted
    pre-main sequence stars
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    https://wiki.sdss.org/pages/viewpage.action?spaceKey=OPS&title=Cartons+for+v1.0
    Additional source catalogs needed: catalogdb.sagitta_edr3
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    apogee_bright_3x1 (for 7 < H < 13)
    Implementation: (in sagitta_edr3) & h<13
    lead contact:Marina Kounkel
    """

    # peewee Model name ---> postgres table name
    # Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
    # Gaia_DR3(CatalogdbModel)--->'gaia_dr3_source'
    # Zari18pms(CatalogdbModel)--->'catalogdb.zari18pms'
    # Zari18ums(CatalogdbModel)--->'catalogdb.zari18ums'
    # Sagitta_EDR3(CatalogdbModel)--->'catalogdb.sagitta_edr3'
    # TwoMassPSC(CatalogdbModel)--->'catalogdb.twomass_psc'

    name = "mwm_yso_pms_apogee_sagitta_edr3"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2700
    can_offset = True

    def build_query(self, version_id, query_region=None):
        # join with Sagitta_EDR3 (we use Gaia_DR3 for query)
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag.alias("gaia_dr3_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR3.parallax,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(Gaia_DR3)
            .join(Sagitta_EDR3, on=(Gaia_DR3.source_id == Sagitta_EDR3.source_id))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < 13,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_YSO_PMS_APOGEE_Sagitta_EDR3_Single_Carton(MWM_YSO_PMS_APOGEE_Sagitta_EDR3_Carton):
    name = "mwm_yso_pms_apogee_sagitta_edr3_single"
    cadence = "bright_1x1"
    priority = 2701  # pms so different from other single cartons


class MWM_YSO_PMS_APOGEE_Zari18pms_Carton(BaseCarton):
    """
    YSOs - Pre-main sequence, APOGEE
    Shorthand name: mwm_yso_pms_apogee_zari18pms
    Comments: NA
    Simplified Description of selection criteria:
    Selecting the clustered sources from the catalog of vetted
    pre-main sequence stars
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    https://wiki.sdss.org/pages/viewpage.action?spaceKey=OPS&title=Cartons+for+v1.0
    Additional source catalogs needed: catalogdb.zari18pms
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    apogee_bright_3x1 (for 7 < H < 13)
    Implementation: (in zari18pms) & h<13
    lead contact:Marina Kounkel
    """

    # peewee Model name ---> postgres table name
    # Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
    # Gaia_DR3(CatalogdbModel)--->'gaia_dr3_source'
    # Zari18pms(CatalogdbModel)--->'catalogdb.zari18pms'
    # Zari18ums(CatalogdbModel)--->'catalogdb.zari18ums'
    # Sagitta_EDR3(CatalogdbModel)--->'catalogdb.sagitta_edr3'
    # TwoMassPSC(CatalogdbModel)--->'catalogdb.twomass_psc'

    name = "mwm_yso_pms_apogee_zari18pms"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_3x1"
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2700
    can_offset = True

    def build_query(self, version_id, query_region=None):
        # join with Zari18pms (we use Gaia_DR2 for query and not Gaia_DR3)
        query = (
            CatalogToTIC_v8.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag.alias("gaia_dr2_rp"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
                Gaia_DR2.parallax,
            )
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .switch(TIC_v8)
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .switch(Gaia_DR2)
            .join(Zari18pms, on=(Gaia_DR2.source_id == Zari18pms.source))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                TwoMassPSC.h_m < 13,
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_YSO_PMS_APOGEE_Zari18pms_Single_Carton(MWM_YSO_PMS_APOGEE_Zari18pms_Carton):
    name = "mwm_yso_pms_apogee_zari18pms_single"
    cadence = "bright_1x1"
    priority = 2701  # pms so different from other single cartons


class MWM_YSO_PMS_BOSS_Sagitta_EDR3_Carton(BaseCarton):
    """
    YSOs - Pre-main sequence, BOSS
    Shorthand name: mwm_yso_pms_boss_sagitta_edr3
    Comments: For v1.0, we use sagitta_edr3 instead of sagitta.
    Simplified Description of selection criteria:
    Selecting the clustered sources from the catalog of vetted
    pre-main sequence stars
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    https://wiki.sdss.org/pages/viewpage.action?spaceKey=OPS&title=Cartons+for+v1.0
    Additional source catalogs needed: catalogdb.sagitta_edr3
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets:
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Implementation: (in sagitta_edr3) & rp<15.5
    lead contact:Marina Kounkel
    """

    # peewee Model name ---> postgres table name
    # Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
    # Gaia_DR3(CatalogdbModel)--->'gaia_dr3_source'
    # Zari18pms(CatalogdbModel)--->'catalogdb.zari18pms'
    # Zari18ums(CatalogdbModel)--->'catalogdb.zari18ums'
    # Sagitta_EDR3(CatalogdbModel)--->'catalogdb.sagitta_edr3'
    # TwoMassPSC(CatalogdbModel)--->'catalogdb.twomass_psc'

    name = "mwm_yso_pms_boss_sagitta_edr3"
    category = "science"
    instrument = None  # instrument is set in post_process()
    cadence = None  # cadence is set in post_process()
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2700
    can_offset = True

    def build_query(self, version_id, query_region=None):
        # join with Sagitta_EDR3 (we use Gaia_DR3 for query)
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag.alias("gaia_dr3_rp"),
                Gaia_DR3.parallax,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(Sagitta_EDR3, on=(Gaia_DR3.source_id == Sagitta_EDR3.source_id))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                Gaia_DR3.phot_rp_mean_mag < 15.5,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr3_rp from " + " sandbox.temp_mwm_yso_pms_boss_sagitta_edr3 ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if current_rp < 14.76:
                current_instrument = "BOSS"
                current_cadence = "bright_3x1"
            elif current_rp < 15.075:
                current_instrument = "BOSS"
                current_cadence = "bright_4x1"
            elif current_rp < 15.29:
                current_instrument = "BOSS"
                current_cadence = "bright_5x1"
            elif current_rp < 15.5:
                current_instrument = "BOSS"
                current_cadence = "bright_6x1"
            else:
                # All cases should be covered above so we should not get here.
                current_instrument = None
                current_cadence = None
                raise TargetSelectionError(
                    "error in mwm_yso_pms_boss_sagitta_edr3 "
                    + "post_process(): "
                    + "instrument = None, cadence= None"
                )

            if current_instrument is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_pms_boss_sagitta_edr3 "
                    + " set instrument = '"
                    + current_instrument
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_pms_boss_sagitta_edr3 "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )


class MWM_YSO_PMS_BOSS_Sagitta_EDR3_Single_Carton(MWM_YSO_PMS_BOSS_Sagitta_EDR3_Carton):
    name = "mwm_yso_pms_boss_sagitta_edr3_single"
    cadence = "bright_1x1"
    priority = 2701  # pms so different from other single cartons
    instrument = "BOSS"

    def post_process(self, model):
        pass


class MWM_YSO_PMS_BOSS_Zari18pms_Carton(BaseCarton):
    """
    YSOs - Pre-main sequence, BOSS
    Shorthand name: mwm_yso_pms_boss_zari18pms
    Comments: NA
    Simplified Description of selection criteria:
    Selecting the clustered sources from the catalog of vetted
    pre-main sequence stars
    Wiki page:
    https://wiki.sdss.org/display/MWM/YSO+selection+function
    https://wiki.sdss.org/pages/viewpage.action?spaceKey=OPS&title=Cartons+for+v1.0
    Additional source catalogs needed: catalogdb.zari18pms
    Return columns: Gaia id, 2mass id, G, BP, RP, J, H, K, parallax
    cadence options for these targets:
    boss_bright_3x1 if RP<14.76 |
    boss_bright_4x1 if RP<15.075 |
    boss_bright_5x1 if RP<15.29 |
    boss_bright_6x1 if RP<15.5
    Implementation: (in zari18pms) & rp<15.5
    lead contact:Marina Kounkel
    """

    # peewee Model name ---> postgres table name
    # Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
    # Gaia_DR3(CatalogdbModel)--->'gaia_dr3_source'
    # Zari18pms(CatalogdbModel)--->'catalogdb.zari18pms'
    # Zari18ums(CatalogdbModel)--->'catalogdb.zari18ums'
    # Sagitta_EDR3(CatalogdbModel)--->'catalogdb.sagitta_edr3'
    # TwoMassPSC(CatalogdbModel)--->'catalogdb.twomass_psc'

    name = "mwm_yso_pms_boss_zari18pms"
    category = "science"
    instrument = None  # instrument is set in post_process()
    cadence = None  # cadence is set in post_process()
    program = "mwm_yso"
    mapper = "MWM"
    priority = 2700
    can_offset = True

    def build_query(self, version_id, query_region=None):
        # join with Zari18pms (we use Gaia_DR2 for query and not Gaia_DR3)
        query = (
            CatalogToTIC_v8.select(
                CatalogToTIC_v8.catalogid,
                Gaia_DR2.source_id,
                Gaia_DR2.ra.alias("gaia_dr2_ra"),
                Gaia_DR2.dec.alias("gaia_dr2_dec"),
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag.alias("gaia_dr2_rp"),
                Gaia_DR2.parallax,
            )
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .switch(TIC_v8)
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .switch(Gaia_DR2)
            .join(Zari18pms, on=(Gaia_DR2.source_id == Zari18pms.source))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                Gaia_DR2.phot_rp_mean_mag < 15.5,
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """
        cadence options for these targets:
        boss_bright_3x1 if RP<14.76 |
        boss_bright_4x1 if RP<15.075 |
        boss_bright_5x1 if RP<15.29 |
        boss_bright_6x1 if RP<15.5
        """

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr2_rp from " + " sandbox.temp_mwm_yso_pms_boss_zari18pms ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_rp = output[i][1]

            if current_rp < 14.76:
                current_instrument = "BOSS"
                current_cadence = "bright_3x1"
            elif current_rp < 15.075:
                current_instrument = "BOSS"
                current_cadence = "bright_4x1"
            elif current_rp < 15.29:
                current_instrument = "BOSS"
                current_cadence = "bright_5x1"
            elif current_rp < 15.5:
                current_instrument = "BOSS"
                current_cadence = "bright_6x1"
            else:
                # All cases should be covered above so we should not get here.
                current_instrument = None
                current_cadence = None
                raise TargetSelectionError(
                    "error in mwm_yso_pms_boss_zari18pms "
                    + "post_process(): "
                    + "instrument = None, cadence= None"
                )

            if current_instrument is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_pms_boss_zari18pms "
                    + " set instrument = '"
                    + current_instrument
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_yso_pms_boss_zari18pms "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )


class MWM_YSO_PMS_BOSS_Zari18pms_Single_Carton(MWM_YSO_PMS_BOSS_Zari18pms_Carton):
    name = "mwm_yso_pms_boss_zari18pms_single"
    cadence = "bright_1x1"
    priority = 2701  # pms so different from other single cartons
    instrument = "BOSS"

    def post_process(self, model):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-04-11
# @Filename: utils.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import contextlib
import hashlib
import io
import time

import click
import pandas
from peewee import SQL, DoesNotExist, ProgrammingError, fn


__all__ = [
    "Timer",
    "sql_apply_pm",
    "sql_iauname",
    "copy_pandas",
    "get_epoch",
    "set_config_parameter",
    "remove_version",
    "vacuum_outputs",
    "get_configuration_values",
    "vacuum_table",
]


class Timer:
    """Convenience context manager to time events.

    Modified from https://bit.ly/3ebdp3y.

    """

    def __enter__(self):
        self.start = time.time()
        self.end = None
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.interval = self.end - self.start

    @property
    def elapsed(self):
        if self.end:
            return self.interval

        return time.time() - self.start


def sql_apply_pm(ra_field, dec_field, pmra_field, pmdec_field, epoch_delta, is_pmra_cos=True):
    """Constructs a SQL expression for applying proper motions to RA/Dec.

    Parameters
    ----------
    ra_field
        The Peewee field or value representing Right Ascension.
    dec_field
        The Peewee field or value representing Declination.
    pmra_field
        The Peewee field or value representing the proper motion in RA.
    pmdec_field
        The Peewee field or value representing the proper motion in Dec.
    epoch_delta
        A value or expression with the delta epoch, in years.
    is_pmra_cos : bool
        Whether the ``pmra_field`` includes the correction for the cosine of
        the declination

    Returns
    -------
    ra_dec : `~peewee:Expression`
        A tuple with the `~peewee:Expression` instances for the proper motion
        corrected RA and Dec.

    """

    pmra_field = fn.coalesce(pmra_field, 0) / 1000.0 / 3600.0
    pmdec_field = fn.coalesce(pmdec_field, 0) / 1000.0 / 3600.0
    epoch = fn.coalesce(epoch_delta, 0)

    if is_pmra_cos:
        cos_dec = fn.cos(fn.radians(fn.coalesce(dec_field, 0)))
        ra_corr = ra_field + pmra_field / cos_dec * epoch
    else:
        ra_corr = ra_field + pmra_field * epoch

    dec_corr = dec_field + pmdec_field * epoch

    return ra_corr, dec_corr


def sql_iauname(ra_field, dec_field, prefix="SDSS J"):
    """Constructs a SQL expression for the IAU name from RA/Dec.

    Parameters
    ----------
    ra_field
        The Peewee field or value representing Right Ascension.
    dec_field
        The Peewee field or value representing Declination.
    prefix : str
        The prefix to add before the calculated IAU name.

    Returns
    -------
    iauname : `~peewee:Expression`
        The `~peewee:Expression` instance to construct the IAU name.

    """

    def dd2dms(dd, round=2, sign=False):
        degs = fn.trunc(dd).cast("INT")
        mins = fn.trunc((fn.abs(dd) - fn.abs(degs)) * 60).cast("INT")

        secs_raw = ((((fn.abs(dd) - fn.abs(degs)) * 60) - mins) * 60).cast("NUMERIC")
        secs = fn.round(secs_raw, round).cast("REAL")

        degs_fmt = "00" if sign is False else "SG00"
        secs_fmt = "00." + "0" * round

        return (
            fn.trim(fn.to_char(degs, degs_fmt))
            .concat(fn.trim(fn.to_char(mins, "00")))
            .concat(fn.trim(fn.to_char(secs, secs_fmt)))
        )

    return (
        SQL("'" + prefix + "'")
        .concat(dd2dms(ra_field / 15))
        .concat(dd2dms(dec_field, round=1, sign=True))
    )


def copy_pandas(df, database, table_name, schema=None, columns=None):
    """Inserts a Pandas data frame using COPY.

    Parameters
    ----------
    df : ~pandas.DataFrame
        The Pandas data frame to copy or the inputs to be used to create one.
    database
        A database connection.
    table_name : str
        The name of the table into which to copy the data.
    schema : str
        The schema in which the table lives.
    columns : list
        The list of column names to copy.

    """

    if not isinstance(df, pandas.DataFrame):
        df = pandas.DataFrame(data=df)

    stream = io.StringIO()
    df.to_csv(stream, index=False, header=False)
    stream.seek(0)

    cursor = database.cursor()

    full_table_name = schema + "." + table_name if schema else table_name

    with database.atomic():
        cursor.copy_from(stream, full_table_name, columns=columns, sep=",")

    return df


def get_epoch(xmodel):
    """Returns the epoch for an `.XMatchModel` in Julian years."""

    xmatch = xmodel._meta.xmatch
    fields = xmodel._meta.fields

    if not xmatch.epoch and not xmatch.epoch_column:
        return None

    # If epoch == 0, make it null. This helps with q3c functions.
    if xmatch.epoch:
        epoch = xmatch.epoch
    else:
        epoch = fn.nullif(fields[xmatch.epoch_column], 0)

    if xmatch.epoch_format == "jd":
        epoch = 2000 + (epoch - 2451545.0) / 365.25

    return epoch


@contextlib.contextmanager
def set_config_parameter(database, parameter, new_value, reset=True, log=None):
    """Temporarily a database configuration parameter."""

    new_value = new_value.upper()
    orig_value = None
    value_changed = None

    try:
        orig_value = database.execute_sql(f"SHOW {parameter}").fetchone()[0].upper()
        value_changed = orig_value != new_value

        if value_changed:
            database.execute_sql(f"SET {parameter} = {new_value};")
            if log:
                log.debug(f"{parameter} is {new_value}.")

        yield

    finally:
        if reset and value_changed:
            database.execute_sql(f"SET {parameter} = {orig_value};")
            if log:
                log.debug(f"{parameter} reset to {orig_value}.")


def remove_version(
    database,
    plan,
    schema="catalogdb",
    table="catalog",
    delete_version=True,
    vacuum=True,
):
    """Removes all rows in ``table`` and ``table_to_`` that match a version."""

    models = []
    for path in database.models:
        schema, table_name = path.split(".")
        if schema != schema:
            continue
        if table_name != table and not table_name.startswith(table + "_to_"):
            continue
        model = database.models[path]
        if not model.table_exists() or model._meta.schema != schema:
            continue
        models.append(model)

    if len(models) == 0:
        raise ValueError("No table matches the input parameters.")

    print(
        f"Tables that will be truncated on {plan!r}: "
        + ", ".join(model._meta.table_name for model in models)
    )

    Catalog = database.models[schema + "." + table]
    Version = database.models[schema + ".version"]

    try:
        version_id = Version.get(plan=plan).id
    except DoesNotExist:
        raise ValueError(f"Version {plan!r} does not exist.")

    print(f"version_id={version_id}")

    n_targets = Catalog.select().where(Catalog.version_id == version_id).count()
    print(f"Number of targets found in {Catalog._meta.table_name}: {n_targets:,}")

    if not click.confirm("Do you really want to proceed?"):
        return

    for model in models:
        n_removed = model.delete().where(model.version_id == version_id).execute()
        print(f"{model._meta.table_name}: {n_removed:,} rows removed.")
        if vacuum:
            print("Vacuuming ...")
            vacuum_table(database, f"{model._meta.schema}.{model._meta.table_name}")

    md5 = hashlib.md5(plan.encode()).hexdigest()[0:16]
    for table in database.get_tables(schema=schema):
        if table.endswith(md5):
            print(f"Dropping temporary table {table}.")
            database.execute_sql(f"DROP TABLE IF EXISTS {table};")

    if delete_version:
        Version.delete().where(Version.id == version_id).execute()
        print("Removed entry in 'version'.")


def vacuum_table(database, table_name, vacuum=True, analyze=True, maintenance_work_mem="50GB"):
    """Vacuums and analyses a table."""

    statement = ("VACUUM " if vacuum else "") + ("ANALYZE " if analyze else "") + table_name

    with database.atomic():
        # Change isolation level to allow executing commands such as VACUUM.
        connection = database.connection()
        original_isolation_level = connection.isolation_level
        connection.set_isolation_level(0)

        database.execute_sql(f"SET maintenance_work_mem = {maintenance_work_mem!r}")

        database.execute_sql(statement)

        connection.set_isolation_level(original_isolation_level)


def vacuum_outputs(
    database,
    vacuum=True,
    analyze=True,
    schema="catalogdb",
    table="catalog",
    relational_tables=True,
):
    """Vacuums and analyses the output tables."""

    assert vacuum or analyze, "either vacuum or analyze need to be True."
    assert database.is_connection_usable(), "connection is not usable."

    tables = []
    for full_name in database.models:
        table_schema, table_name = full_name.split(".")
        if table_schema != schema:
            continue
        if table_name != table:
            is_relational = table_name.startswith(table + "_to_")
            if not is_relational or not relational_tables:
                continue
        tables.append(table_name)

    for table_name in tables:
        table_name = table_name if schema is None else schema + "." + table_name
        vacuum_table(database, table_name, vacuum=vacuum, analyze=analyze)


def get_configuration_values(database, parameters):
    """Returns a dictionary of datbase configuration parameter to value."""

    values = {}

    with database.atomic():
        for parameter in parameters:
            try:
                value = database.execute_sql(f"SHOW {parameter}").fetchone()[0]
                values[parameter] = value
            except ProgrammingError:
                database.rollback()

    return values


def is_view(database, view_name, schema="public", materialized=False):
    """Determines if a view exists."""

    if not materialized:
        query = (
            f"SELECT * FROM pg_views where schemaname = {schema!r} "
            f"AND viewname = {view_name!r};"
        )
    else:
        query = (
            f"SELECT pc.* FROM pg_class pc "
            f"JOIN pg_namespace pn ON pc.relnamespace = pn.oid "
            f"WHERE pn.nspname = {schema!r} AND "
            f"pc.relname = {view_name!r};"
        )

    if len(database.execute_sql(query).fetchall()) > 0:
        return True
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-06-12
# @Filename: bhm_csc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import JOIN, fn

from sdssdb.peewee.sdss5db.catalogdb import (
    BHM_CSC_v3,
    CatalogFromSDSS_DR19p_Speclite,
    CatalogToGaia_DR3,
    CatalogToLegacy_Survey_DR10,
    CatalogToPanstarrs1,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    Legacy_Survey_DR10,
    Panstarrs1,
    SDSS_DR19p_Speclite,
    TwoMassPSC,
)

from target_selection.cartons.base import BaseCarton
from target_selection.mag_flux import AB2Jy, AB2nMgy


# DEBUG STUFF TO USE TEMP TABLE
# CatalogToSDSS_DR19p_Speclite._meta.table_name = 'temp_catalog_to_sdss_dr19p_speclite'
# CatalogToSDSS_DR19p_Speclite._meta._schema = 'sandbox'

# Details: Start here
# https://wiki.sdss.org/display/OPS/Cartons+for+v0.5#Cartonsforv0.5-BHMCSC # noqa: E501

# This module provides the following BHM cartons:
#   bhm_csc_boss
#   bhm_csc_boss_d3
#   bhm_csc_apogee


class BhmCscBossCarton(BaseCarton):
    """ """

    name = "bhm_csc_boss"
    category = "science"
    mapper = "BHM"
    program = "bhm_csc"
    instrument = "BOSS"
    tile = False
    can_offset = True
    only_faintest_cadence = False

    def build_query(self, version_id, query_region=None):
        x = BHM_CSC_v3.alias()
        ps = Panstarrs1.alias()
        c2ps = CatalogToPanstarrs1.alias()
        g3 = Gaia_DR3.alias()
        c2g3 = CatalogToGaia_DR3.alias()
        ls = Legacy_Survey_DR10.alias()
        c2ls = CatalogToLegacy_Survey_DR10.alias()

        g_psf_flux_max = AB2Jy(self.parameters["g_psf_mag_min"])
        r_psf_flux_max = AB2Jy(self.parameters["r_psf_mag_min"])
        i_psf_flux_max = AB2Jy(self.parameters["i_psf_mag_min"])
        z_psf_flux_max = AB2Jy(self.parameters["z_psf_mag_min"])
        g_psf_flux_min = AB2Jy(self.parameters["g_psf_mag_max"])
        r_psf_flux_min = AB2Jy(self.parameters["r_psf_mag_max"])
        i_psf_flux_min = AB2Jy(self.parameters["i_psf_mag_max"])
        z_psf_flux_min = AB2Jy(self.parameters["z_psf_mag_max"])
        fiberflux_g_max = AB2nMgy(self.parameters["fibermag_g_min"])
        fiberflux_r_max = AB2nMgy(self.parameters["fibermag_r_min"])
        fiberflux_i_max = AB2nMgy(self.parameters["fibermag_i_min"])
        fiberflux_z_max = AB2nMgy(self.parameters["fibermag_z_min"])
        fiberflux_g_min = AB2nMgy(self.parameters["fibermag_g_max"])
        fiberflux_r_min = AB2nMgy(self.parameters["fibermag_r_max"])
        fiberflux_i_min = AB2nMgy(self.parameters["fibermag_i_max"])
        fiberflux_z_min = AB2nMgy(self.parameters["fibermag_z_max"])

        gaia_g_mag_min = self.parameters["gaia_g_mag_min"]
        gaia_g_mag_max = self.parameters["gaia_g_mag_max"]

        i_psf_flux_min_for_cadence1 = AB2Jy(self.parameters["i_psf_mag_max_for_cadence1"])
        fiberflux_r_min_for_cadence1 = AB2nMgy(self.parameters["fibermag_r_max_for_cadence1"])
        gaia_g_mag_max_for_cadence1 = self.parameters["gaia_g_mag_max_for_cadence1"]
        i_psf_flux_min_for_cadence2 = AB2Jy(self.parameters["i_psf_mag_max_for_cadence2"])
        fiberflux_r_min_for_cadence2 = AB2nMgy(self.parameters["fibermag_r_max_for_cadence2"])
        gaia_g_mag_max_for_cadence2 = self.parameters["gaia_g_mag_max_for_cadence2"]

        value = peewee.Value(self.parameters["value"]).cast("real")

        cadence1 = self.parameters["cadence1"]
        cadence2 = self.parameters["cadence2"]
        cadence3 = self.parameters["cadence3"]
        cadence = peewee.Case(
            None,
            (
                (
                    (
                        (ps.i_stk_psf_flux > i_psf_flux_min_for_cadence1)
                        | (ls.fiberflux_r > fiberflux_r_min_for_cadence1)
                        | (g3.phot_g_mean_mag < gaia_g_mag_max_for_cadence1)
                    ),
                    cadence1,
                ),
                (
                    (
                        (ps.i_stk_psf_flux > i_psf_flux_min_for_cadence2)
                        | (ls.fiberflux_r > fiberflux_r_min_for_cadence2)
                        | (g3.phot_g_mean_mag < gaia_g_mag_max_for_cadence2)
                    ),
                    cadence2,
                ),
            ),
            cadence3,
        )

        # #########################################################################
        # prepare the existing spectroscopy catalogues - these affect only priority
        spec_sn_thresh = self.parameters["spec_sn_thresh"]
        spec_z_err_thresh = self.parameters["spec_z_err_thresh"]
        dpriority_has_spec = self.parameters["dpriority_has_spec"]

        # SDSS DR19p
        # downslect only 'good' spectra
        c2s19 = CatalogFromSDSS_DR19p_Speclite.alias()
        ss19 = SDSS_DR19p_Speclite.alias()
        s19 = (
            ss19.select(
                ss19.pk.alias("s19_pk"),
            )
            .where(
                ss19.sn_median_all >= spec_sn_thresh,
                ss19.zwarning == 0,
                ss19.z_err <= spec_z_err_thresh,
                ss19.z_err > 0.0,
                ss19.specprimary > 0,
            )
            .alias("s19")
        )

        # adjust priority if target aleady has a good SDSS spectrum
        priority_1 = peewee.Case(None, ((s19.c.s19_pk.is_null(False), 1),), 0)
        #
        # Compute net priority
        priority_floor_dark = peewee.Value(self.parameters["priority_floor_dark"])
        priority_floor_bright = peewee.Value(self.parameters["priority_floor_bright"])
        priority = peewee.Case(
            None,
            (
                (
                    (
                        (ps.i_stk_psf_flux > i_psf_flux_min_for_cadence1)
                        | (g3.phot_g_mean_mag < gaia_g_mag_max_for_cadence1)
                    ),
                    priority_floor_bright + (priority_1 * dpriority_has_spec) + x.xpriority - 1,
                ),
            ),
            priority_floor_dark + (priority_1 * dpriority_has_spec) + x.xpriority - 1,
        )

        query = (
            x.select(
                fn.coalesce(c2g3.catalogid, c2ls.catalogid, c2ps.catalogid).alias("catalogid"),
                priority.alias("priority"),
                cadence.alias("cadence"),
                value.alias("value"),
                x.pk.alias("csc_pk"),  # extra
                x.csc21p_id.alias("csc_csc21p_id"),  # extra
                x.ra.alias("csc_opt_ra"),  # extra
                x.dec.alias("csc_opt_dec"),  # extra
                x.best_mag.alias("csc_best_mag"),  # extra
                x.mag_type.alias("csc_mag_type"),  # extra
                x.best_oir_cat.alias("csc_best_oir_cat"),  # extra
                x.xpriority.alias("csc_xpriority"),  # extra
                x.xband.alias("csc_xband"),  # extra
                x.logfx.alias("csc_logfx"),  # extra
                x.gaia_dr3_srcid.alias("csc_gaia_dr3_srcid"),  # extra
                x.ls_dr10_lsid.alias("csc_ls_dr10_lsid"),  # extra
                x.ps21p_ippobjid.alias("csc_ps21p_ippobjid"),  # extra
                g3.phot_g_mean_mag.alias("g3_g_mag"),  # extra
                ls.fiberflux_r.alias("ls_fiberflux_r"),  # extra
                ls.fiberflux_i.alias("ls_fiberflux_i"),  # extra
                ps.r_stk_psf_flux.alias("ps_r_stk_psf_flux"),  # extra
                ps.i_stk_psf_flux.alias("ps_i_stk_psf_flux"),  # extra
                c2s19.target_id.alias("sdss_dr19p_speclite_pk"),  # extra
            )
            .join(g3, join_type=JOIN.LEFT_OUTER, on=(x.gaia_dr3_srcid == g3.source_id))
            .join(ls, join_type=JOIN.LEFT_OUTER, on=(x.ls_dr10_lsid == ls.ls_id))
            .join(ps, join_type=JOIN.LEFT_OUTER, on=(x.ps21p_ippobjid == ps.catid_objid))
            .join(c2ps, join_type=JOIN.LEFT_OUTER, on=(ps.catid_objid == c2ps.target_id))
            .join(c2ls, join_type=JOIN.LEFT_OUTER, on=(ls.ls_id == c2ls.target_id))
            .join(c2g3, join_type=JOIN.LEFT_OUTER, on=(g3.source_id == c2g3.target_id))
            .join(
                c2s19,
                join_type=JOIN.LEFT_OUTER,
                on=(
                    fn.coalesce(c2g3.catalogid, c2ls.catalogid, c2ps.catalogid) == c2s19.catalogid
                ),
            )
            .join(s19, join_type=JOIN.LEFT_OUTER, on=(s19.c.s19_pk == c2s19.target_id))
            .where(
                fn.coalesce(c2g3.version_id, c2ls.version_id, c2ps.version_id) == version_id,
                fn.coalesce(c2s19.version_id, version_id) == version_id,
                fn.coalesce(c2g3.best, c2ls.best, c2ps.best) >> True,
            )
            .where(
                (
                    (x.best_oir_cat == "gdr3")
                    & (g3.phot_g_mean_mag.between(gaia_g_mag_min, gaia_g_mag_max))
                )
                | (
                    (x.best_oir_cat == "lsdr10")
                    & (ls.fiberflux_g < fiberflux_g_max)
                    & (ls.fiberflux_r < fiberflux_r_max)
                    & (ls.fiberflux_i < fiberflux_i_max)
                    & (ls.fiberflux_z < fiberflux_z_max)
                    & (
                        (ls.fiberflux_g > fiberflux_g_min)
                        | (ls.fiberflux_r > fiberflux_r_min)
                        | (ls.fiberflux_i > fiberflux_i_min)
                        | (ls.fiberflux_z > fiberflux_z_min)
                    )
                )
                | (
                    (x.best_oir_cat == "ps1dr2")
                    & (ps.g_stk_psf_flux < g_psf_flux_max)
                    & (ps.r_stk_psf_flux < r_psf_flux_max)
                    & (ps.i_stk_psf_flux < i_psf_flux_max)
                    & (ps.z_stk_psf_flux < z_psf_flux_max)
                    & (
                        (ps.g_stk_psf_flux > g_psf_flux_min)
                        | (ps.r_stk_psf_flux > r_psf_flux_min)
                        | (ps.i_stk_psf_flux > i_psf_flux_min)
                        | (ps.z_stk_psf_flux > z_psf_flux_min)
                    )
                )
            )
            # .distinct(fn.coalesce(fn.max(c2ps.catalogid), fn.max(c2g3.catalogid)))
            .distinct([fn.coalesce(c2g3.catalogid, c2ls.catalogid, c2ps.catalogid)])
        )

        if self.only_faintest_cadence:
            query = query.where(cadence == cadence3)

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    x.ora, x.odec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


class BhmCscBossD3Carton(BhmCscBossCarton):
    name = "bhm_csc_boss_d3"
    only_faintest_cadence = True


# #######
class BhmCscApogeeCarton(BaseCarton):
    """
    SELECT * from bhm_csc_v2 AS x
    WHERE
      AND x.best_mag BETWEEN 7 AND 14
    """

    name = "bhm_csc_apogee"
    mapper = "BHM"
    program = "bhm_csc"
    category = "science"
    this_cadence = "bright_3x1"  # TODO Check this is still a valid choice
    instrument = "APOGEE"
    tile = False
    can_offset = True

    def build_query(self, version_id, query_region=None):
        x = BHM_CSC_v3.alias()
        # tic = TIC_v8.alias()
        # c2tic = CatalogToTIC_v8.alias()
        # change to now rely directly on twomass_psc table instead of TIC_v8
        tm = TwoMassPSC.alias()
        c2tm = CatalogToTwoMassPSC.alias()

        hmag_max_for_cadence1 = self.parameters["hmag_max_for_cadence1"]

        value = peewee.Value(self.parameters["value"]).cast("real")

        cadence = peewee.Case(
            None,
            (((x.best_mag < hmag_max_for_cadence1), self.parameters["cadence1"]),),
            self.parameters["cadence2"],
        )

        # Compute net priority
        priority = peewee.Value(self.parameters["priority_floor"]) + x.xpriority - 1

        query = (
            x.select(
                # c2tic.catalogid.alias('catalogid'),
                c2tm.catalogid.alias("catalogid"),
                x.csc21p_id.alias("csc_csc21p_id"),  # extra
                priority.alias("priority"),
                cadence.alias("cadence"),
                value.alias("value"),
                tm.j_m.alias("j"),
                tm.h_m.alias("h"),
                tm.k_m.alias("k"),
                tm.pts_key.alias("twomass_pts_key"),
                x.best_oir_cat.alias("csc_best_oir_cat"),  # extra
                x.ra.alias("csc_oir_ra"),  # extra
                x.dec.alias("csc_oir_dec"),  # extra
                x.best_mag.alias("csc_best_mag"),  # extra
                x.mag_type.alias("csc_mag_type"),  # extra
                x.xpriority.alias("csc_xpriority"),  # extra
                x.xband.alias("csc_xband"),  # extra
                x.logfx.alias("csc_logfx"),  # extra
                x.tmass_designation.alias("csc_tmass_designation"),  # extra
            )
            .join(tm, on=(x.tmass_designation == tm.designation))
            .join(c2tm, on=(tm.pts_key == c2tm.target_id))
            .where(
                c2tm.version_id == version_id,
                c2tm.best >> True,
                # x.best_oir_cat == '2mass',
                # x.best_mag >= self.parameters['hmag_min'],
                # x.best_mag < self.parameters['hmag_max'],
                # x.best_mag != 'NaN',
                tm.h_m.between(self.parameters["hmag_min"], self.parameters["hmag_max"]),
            )
            .distinct([c2tm.catalogid])
        )

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    x.ra2m, x.dec2m, query_region[0], query_region[1], query_region[2]
                )
            )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# Updated Nov 2021 by Tom Dwelly and Felipe Santana
# @Date: 2020-07-26
# @Filename: ops_skies.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import Model, fn

from sdssdb.peewee.sdss5db import database
from sdssdb.peewee.sdss5db.catalogdb import CatalogToSkies_v2, Skies_v2

from target_selection.cartons.base import BaseCarton


class OPS_Sky_Boss_Best_Carton(BaseCarton):
    """Top quality skies for the BOSS spectrograph.

    Definition:
        Select sky positions from catalogdb.skies_v2 that don't have a
        nearby Gaia, Tycho2, or 2MASS source and that are
        valid in at least lsdr8 OR ps1dr2 (and not obviously invalid in the other).
        Take care near edges of ls8 and ps1dr2 footprints by selecting
        only sky locations that have at least one neighbour (not too close)
        in their originating catalogue.

    """

    name = "ops_sky_boss_best"
    cadence = None
    category = "sky_boss"
    program = "ops_sky"
    mapper = None
    instrument = "BOSS"
    inertial = True
    priority = 5000
    can_offset = False

    load_magnitudes = False

    def build_query(self, version_id, query_region=None):
        a_large_value = 1e30
        pars = self.parameters

        query = (
            Skies_v2.select(
                CatalogToSkies_v2.catalogid,
                Skies_v2.ra,
                Skies_v2.dec,
                Skies_v2.pix_32768,
                Skies_v2.tile_32,
            )
            .join(CatalogToSkies_v2)
            .where(CatalogToSkies_v2.version_id == version_id, CatalogToSkies_v2.best >> True)
            .where(
                Skies_v2.valid_gaia >> True,
                Skies_v2.valid_tmass >> True,
                Skies_v2.valid_tycho2 >> True,
                Skies_v2.valid_tmass_xsc >> True,
            )
            .where(
                fn.COALESCE(Skies_v2.sep_neighbour_ls8, a_large_value) > pars["min_sep_ls8"],
                fn.COALESCE(Skies_v2.sep_neighbour_ps1dr2, a_large_value) > pars["min_sep_ps1dr2"],
                ((Skies_v2.valid_ls8 >> True) & (Skies_v2.sep_neighbour_ls8 < pars["max_sep_ls8"]))
                | (
                    (Skies_v2.valid_ps1dr2 >> True)
                    & (Skies_v2.sep_neighbour_ps1dr2 < pars["max_sep_ps1dr2"])
                ),
            )
        )

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Skies_v2.ra, Skies_v2.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


class OPS_Sky_Boss_Good_Carton(BaseCarton):
    """Reasonable quality skies for the BOSS spectrograph.
       Use these skies when there are not enough OPS_Sky_Boss_Best_Carton skies available

    Definition:
        Select sky positions from catalogdb.skies_v2 that don't have a
        nearby Gaia, Tycho2, or 2MASS source.
        Results in a catalogue that covers the vast majority of the sky,
        excluding only the Galactic bulge and inner parts of the Magellanic clouds.

    """

    name = "ops_sky_boss_good"
    cadence = None
    category = "sky_boss"
    program = "ops_sky"
    mapper = None
    instrument = "BOSS"
    inertial = True
    priority = 5001
    can_offset = False

    load_magnitudes = False

    def build_query(self, version_id, query_region=None):
        pars = self.parameters

        query = (
            Skies_v2.select(
                CatalogToSkies_v2.catalogid,
                Skies_v2.ra,
                Skies_v2.dec,
                Skies_v2.pix_32768,
                Skies_v2.tile_32,
            )
            .join(CatalogToSkies_v2)
            .where(Skies_v2.sep_neighbour_gaia > pars["min_sep_gaia"])
            .where(CatalogToSkies_v2.version_id == version_id, CatalogToSkies_v2.best >> True)
            .where(
                Skies_v2.valid_gaia >> True,
                Skies_v2.valid_tmass >> True,
                Skies_v2.valid_tycho2 >> True,
                Skies_v2.valid_tmass_xsc >> True,
            )
        )

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Skies_v2.ra, Skies_v2.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


class TempTableFallbackCarton(Model):
    """The peewee model class TempTableFallbackCarton is used below in the
    carton class OPS_Sky_Boss_Fallback_Carton.
    """

    tile_32 = peewee.IntegerField()
    nsky = peewee.BigIntegerField()

    class Meta:
        schema = "sandbox"
        database = database
        table_name = "temp_ops_sky_boss_good_missing_pix_xyz"


class OPS_Sky_Boss_Fallback_Carton(BaseCarton):
    """Unconstrained skies for the BOSS spectrograph.
       Use these skies when there are not enough OPS_Sky_Boss_Best_Carton,
       or OPS_Sky_Boss_Good_Carton skies available

    Definition:
       Select tile_32 locations that have few best/good skies available
       Take all sky locations in those pixels

    """

    name = "ops_sky_boss_fallback"
    cadence = None
    category = "sky_boss"
    program = "ops_sky"
    mapper = None
    instrument = "BOSS"
    inertial = True
    priority = 5002
    can_offset = False

    load_magnitudes = False

    """
    Here is the SQL to generate the query.

# Algorithm:
# 1) First find all nside=32 pixels with fewer than 1000 skies/pix in ops_sky_boss_good
#    Here I use the temp table, but would be better to read carton from targetdb or
#    just to re-issue the ops_sky_boss_good query
#    There are 12288 healpixels in NSIDE=32, and so we do the
#    generate_series(0,12287) step to make sure we catch them all
#
# 2) Now select sky candidates (from skies_v2) that land in those pixels,
#    and filter them with some less rigorous criteria than used for good+best cartons

DROP TABLE IF EXISTS sandbox.temp_ops_sky_boss_good_missing_pix ;

SELECT p.tile_32,COALESCE(b.nsky,0) as nsky
INTO sandbox.temp_ops_sky_boss_good_missing_pix
FROM (SELECT generate_series(0,12287) AS tile_32) AS p
     LEFT OUTER JOIN
     (SELECT tile_32,count(*) AS nsky FROM sandbox.temp_ops_sky_boss_good GROUP BY tile_32) AS b
     ON p.tile_32 = b.tile_32
     WHERE COALESCE(b.nsky,0) < 1000 ;

CREATE INDEX ON sandbox.temp_ops_sky_boss_good_missing_pix (tile_32);
ANALYZE sandbox.temp_ops_sky_boss_good_missing_pix;
DROP TABLE IF EXISTS sandbox.temp_ops_sky_boss_good_missing_pix_skies;

SELECT p.nsky,s.*
INTO sandbox.temp_ops_sky_boss_good_missing_pix_skies
FROM sandbox.temp_ops_sky_boss_good_missing_pix AS p
JOIN skies_v2 AS s
ON p.tile_32 = s.tile_32
WHERE selected_gaia is true
  AND COALESCE(sep_neighbour_gaia,1e30) > 3.0
  AND COALESCE(sep_neighbour_ps1dr2,1e30) > 3.0
  AND COALESCE(sep_neighbour_tycho2,1e30) > 15.0
  AND COALESCE(sep_neighbour_tmass,1e30) > 5.0;

    """

    def build_query(self, version_id, query_region=None):
        pars = self.parameters

        # The below lines ensure that the
        # input parameters for the SQL query have the proper types.
        local_min_sep_gaia = float(pars["min_sep_gaia"])
        local_version_id = int(version_id)

        # We have put commit() after the 'drop table' and 'select into' commands
        # to ensure that the previous command is committed
        # before the next command starts.

        # Above comment uses the name sandbox.temp_ops_sky_boss_good_missing_pix.
        # However, below we are using the name
        # sandbox.temp_ops_sky_boss_good_missing_pix_xyz.
        cursor = self.database.execute_sql(
            "DROP TABLE IF EXISTS sandbox.temp_ops_sky_boss_good_missing_pix_xyz ;"
        )

        # Above comment uses the name sandbox.temp_ops_sky_boss_good.
        # However, sandbox.temp_ops_sky_boss_good is produced by the
        # ops_sky_boss_good carton.
        # Hence, below we are using the name sandbox.temp_ops_sky_boss_good_xyz.
        cursor = self.database.execute_sql(
            "DROP TABLE IF EXISTS sandbox.temp_ops_sky_boss_good_xyz ;"
        )
        self.database.commit()

        # This query is from the ops_sky_boss_good carton.
        cursor = self.database.execute_sql(
            "select ct.catalogid, "
            + "sk.ra, sk.dec, sk.pix_32768, sk.tile_32 "
            + "into sandbox.temp_ops_sky_boss_good_xyz "
            + "from catalogdb.skies_v2 as sk, catalogdb.catalog_to_skies_v2 as ct "
            + "where sk.pix_32768 = ct.target_id and "
            + "sk.sep_neighbour_gaia > "
            + str(local_min_sep_gaia)
            + " and "
            + "ct.version_id = "
            + str(local_version_id)
            + " and "
            + "ct.best = true and "
            + "sk.valid_gaia = true and "
            + "sk.valid_tmass = true and "
            + "sk.valid_tycho2 = true and "
            + "sk.valid_tmass_xsc = true ;"
        )
        self.database.commit()

        cursor = self.database.execute_sql(
            "SELECT p.tile_32,COALESCE(b.nsky,0) as nsky "
            + "INTO sandbox.temp_ops_sky_boss_good_missing_pix_xyz "
            + "FROM (SELECT generate_series(0,12287) AS tile_32) AS p "
            + "LEFT OUTER JOIN "
            + "(SELECT tile_32,count(*) AS nsky "
            + "FROM sandbox.temp_ops_sky_boss_good_xyz GROUP BY tile_32) AS b "
            + "ON p.tile_32 = b.tile_32 "
            + "WHERE COALESCE(b.nsky,0) < 1000 ;"
        )

        cursor = self.database.execute_sql(
            "CREATE INDEX ON sandbox.temp_ops_sky_boss_good_missing_pix_xyz (tile_32);"
        )

        cursor = self.database.execute_sql(  # noqa: F841
            "ANALYZE sandbox.temp_ops_sky_boss_good_missing_pix_xyz;"
        )
        self.database.commit()

        # We do not use the table
        # sandbox.temp_ops_sky_boss_good_missing_pix_skies in this carton.
        # Hence, we do not issue the below drop table from the example SQL code above
        #           DROP TABLE IF EXISTS sandbox.temp_ops_sky_boss_good_missing_pix_skies;
        #
        # Below SQL query is implemented as a peewee query after this comment.
        #        cursor = self.database.execute_sql(
        #            "SELECT p.nsky,s.*" +
        #            "INTO sandbox.temp_ops_sky_boss_good_missing_pix_skies " +
        #            "FROM sandbox.temp_ops_sky_boss_good_missing_pix AS p " +
        #            "JOIN skies_v2 AS s " +
        #            "ON p.tile_32 = s.tile_32 " +
        #            "WHERE selected_gaia is true " +
        #            "AND COALESCE(sep_neighbour_gaia,1e30) > 3.0 " +
        #            "AND COALESCE(sep_neighbour_ps1dr2,1e30) > 3.0 " +
        #            "AND COALESCE(sep_neighbour_tycho2,1e30) > 15.0 " +
        #            "AND COALESCE(sep_neighbour_tmass,1e30) > 5.0;")

        # The peewee model TempTableFallbackCarton corresponds to
        # the table sandbox.temp_ops_sky_boss_good_missing_pix_xyz.
        #

        query = (
            CatalogToSkies_v2.select(
                CatalogToSkies_v2.catalogid,
                Skies_v2.ra,
                Skies_v2.dec,
                Skies_v2.pix_32768,
                Skies_v2.tile_32,
                TempTableFallbackCarton.nsky,
            )
            .join(Skies_v2, on=(CatalogToSkies_v2.target_id == Skies_v2.pix_32768))
            .join(
                TempTableFallbackCarton, on=(Skies_v2.tile_32 == TempTableFallbackCarton.tile_32)
            )
            .where(CatalogToSkies_v2.version_id == version_id, CatalogToSkies_v2.best >> True)
            .where(
                Skies_v2.selected_gaia >> True,
                fn.coalesce(Skies_v2.sep_neighbour_gaia, 1e30) > 3.0,
                fn.coalesce(Skies_v2.sep_neighbour_ps1dr2, 1e30) > 3.0,
                fn.coalesce(Skies_v2.sep_neighbour_tycho2, 1e30) > 15.0,
                fn.coalesce(Skies_v2.sep_neighbour_tmass, 1e30) > 5.0,
            )
        )

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Skies_v2.ra, Skies_v2.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


class OPS_Sky_APOGEE_Best_Carton(BaseCarton):
    """First option of skies for the APOGEE spectrograph.

    Definition:
        Select sky positions from catalogdb.skies_v2 that don't have a
        nearby Gaia, Tycho2, 2MASS or 2MASS_XSC source.

    """

    name = "ops_sky_apogee_best"
    cadence = None
    category = "sky_apogee"
    program = "ops_sky"
    mapper = None
    instrument = "APOGEE"
    inertial = True
    priority = 5200
    can_offset = False

    load_magnitudes = False

    def build_query(self, version_id, query_region=None):
        query = (
            Skies_v2.select(
                CatalogToSkies_v2.catalogid,
                Skies_v2.ra,
                Skies_v2.dec,
                Skies_v2.pix_32768,
                Skies_v2.tile_32,
            )
            .join(CatalogToSkies_v2)
            .where(CatalogToSkies_v2.version_id == version_id, CatalogToSkies_v2.best >> True)
            .where(
                Skies_v2.valid_gaia >> True,
                Skies_v2.valid_tmass >> True,
                Skies_v2.valid_tycho2 >> True,
                Skies_v2.valid_tmass_xsc >> True,
            )
        )

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Skies_v2.ra, Skies_v2.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


class OPS_Sky_APOGEE_Good_Carton(BaseCarton):
    """Last resource skies for the APOGEE spectrograph to be used when carton 1 is not available.

    Definition:
        Select sky positions from catalogdb.skies_v2 that
        are invalid in 2MASS but selected.

    """

    name = "ops_sky_apogee_good"
    cadence = None
    category = "sky_apogee"
    program = "ops_sky"
    mapper = None
    instrument = "APOGEE"
    inertial = True
    priority = 5201
    can_offset = False

    load_magnitudes = False

    def build_query(self, version_id, query_region=None):
        query = (
            Skies_v2.select(
                CatalogToSkies_v2.catalogid,
                Skies_v2.ra,
                Skies_v2.dec,
                Skies_v2.pix_32768,
                Skies_v2.tile_32,
            )
            .join(CatalogToSkies_v2)
            .where(CatalogToSkies_v2.version_id == version_id, CatalogToSkies_v2.best >> True)
            .where(Skies_v2.selected_tmass >> True, Skies_v2.valid_tmass >> False)
            .where(Skies_v2.mag_neighbour_tmass > 10.0)
        )
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Skies_v2.ra, Skies_v2.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2022-12-11
# @Filename: mwm_legacy_ir2opt.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    SDSS_DR17_APOGEE_Allstarmerge,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# peewee model : postgres table name
# Catalog : catalogdb.catalog
# CatalogToTIC_v8 : catalogdb.catalog_to_tic_v8
# Gaia_DR2 : catalogdb.gaia_dr2_source
# SDSS_DR17_APOGEE_Allstarmerge : catalogdb.sdss_dr17_apogee_allstarmerge
# TIC_v8 : catalogdb.tic_v8
# TwoMassPSC : catalogdb.twomass_psc
#
#
# We associate each APOGEE target with a 2MASS source since the
# vast majority of APOGEE targets were selected from the 2MASS PSC.
# APOGEE_ID is the same as
# the “designation” column of the 2MASS PSC;
#
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# For sdss_dr17_apogee_allstarmerge we
# have to strip off the “2M” in the APOGEE_ID.
# However, some apogee_id values do not have 2M in the front.
# (old: for sdss_apogeeallstarmerge_r13 we also had to strip off the
#  “2M” in the APOGEE_ID.)
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# sdss5db=# select apogee_id from catalogdb.SDSS_DR17_APOGEE_Allstarmerge
# where apogee_id like '2M%';
#      apogee_id
# ---------------------
# 2M00000002+7417074
# 2M00000019-1924498
# etc.
# sdss5db=# select apogee_id from catalogdb.SDSS_DR17_APOGEE_Allstarmerge
# where apogee_id not like '2M%';
#    apogee_id
# --------------------
# 19140272-1554055
# 19155129-1617591
#
#
# Historical Note:
# The v0.5 version of this carton used catalogdb.sdss_apogeeallstarmerge_r13.
#


class MWM_Legacy_ir2opt_boss_Carton(BaseCarton):
    """
    Shorthand name: mwm_legacy_ir2opt_boss

    APOGEE targets to be observed with BOSS fibres.
    To be used as a filler for MWM-led plates.

    Simplified Description of selection criteria: Select all Gaia targets that
    have an APOGEE counterpart (i.e., have an entry in sdss_dr17_apogee_allstarmerge)
    with 13 < G < 18, BP > 13, and RP > 13.

    Wiki page: NA

    Additional source catalogs needed: Gaia DR2, sdss_dr17_apogee_allstarmerge

    Additional cross-matching needed: NA

    Return columns: catalog_id, source_id, apogee_id, phot_g_mean_mag,
    phot_rp_mean_mag, phot_bp_mean_mag
    """

    name = "mwm_legacy_ir2opt_boss"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_1x1"
    program = "mwm_legacy"
    mapper = "MWM"
    priority = 6100
    can_offset = True

    # In the below query, we use replace() instead of ltrim() since
    # ltrim('2M20', '2M') will also trim the second 2.

    def build_query(self, version_id, query_region=None):
        query = (
            Catalog.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag,
                TwoMassPSC.pts_key,
                TwoMassPSC.designation,
            )
            .join(CatalogToGaia_DR3, on=(Catalog.catalogid == CatalogToGaia_DR3.catalogid))
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .join(
                SDSS_DR17_APOGEE_Allstarmerge,
                on=(
                    TwoMassPSC.designation
                    == peewee.fn.replace(SDSS_DR17_APOGEE_Allstarmerge.apogee_id, "2M", "")
                ),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                Gaia_DR3.phot_g_mean_mag.between(8, 18),
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query



#!/usr/bin/env python
# encoding: utf-8
#
# @Author: Tom Dwelly
# @Date: Oct-2019
# @Filename: mag_flux.py
# @License: BSD 3-Clause
# @Copyright: Tom Dwelly

from __future__ import absolute_import, division, print_function, unicode_literals

from math import log10

import numpy as np


__invalid_mag_val = 99.9999
__invalid_mag_thresh = 90.0
""" see http://www.sdss3.org/dr8/algorithms/magnitudes.php """


# https://svn.sdss.org/public/repo/eboss/idlspec2d/trunk/pro/spec2d/readplugmap.pro
__psfflux_to_fiber2flux = {
    "u": (1.0 / 2.085),
    "g": (1.0 / 2.085),
    "r": (1.0 / 2.116),
    "i": (1.0 / 2.134),
    "z": (1.0 / 2.135),
}


def fluxmag(flux, zp=None):
    assert zp is not None
    if np.ndim(flux) > 0:
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.nan_to_num(
                zp - 2.5 * np.log10(flux),
                copy=False,
                nan=__invalid_mag_val,
                posinf=__invalid_mag_val,
                neginf=__invalid_mag_val,
            )
    else:
        if flux <= 0.0:
            return __invalid_mag_val
        else:
            return zp - 2.5 * log10(flux)


def mag2flux(mag, zp=None):
    assert zp is not None
    if np.ndim(mag) > 0:
        with np.errstate(divide="ignore", invalid="ignore"):
            flux = np.where(mag >= __invalid_mag_thresh, 0.0, 10.0 ** (-0.4 * (mag - zp)))
    else:
        if mag >= __invalid_mag_thresh:
            flux = 0.0
        else:
            flux = 10.0 ** (-0.4 * (mag - zp))
    return flux


def psfflux2fiber2flux(psfflux, filt=None):
    try:
        c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception(f"Unknown filter: {filt}")
    return psfflux * c


def fiber2flux2psfflux(fiber2flux, filt=None):
    try:
        c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception(f"Unknown filter: {filt}")
    return fiber2flux / c


def psfmag2fiber2mag(psfmag, filt=None):
    try:
        c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception(f"Unknown filter: {filt}")
    return psfmag - (2.5 * log10(c))


def fiber2mag2psfmag(fiber2mag, filt=None):
    try:
        c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception(f"Unknown filter: {filt}")
    return fiber2mag + (2.5 * log10(c))


def psfmag_minus_fiber2mag(filt=None):
    try:
        c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception(f"Unknown filter: {filt}")
    return 2.5 * log10(c)


def fiber2mag_minus_psfmag(filt=None):
    try:
        c = __psfflux_to_fiber2flux[filt]
    except:
        raise Exception(f"Unknown filter: {filt}")
    return -2.5 * log10(c)


# def nMgy2AB(flux):
#     return flux2mag(mag, 22.50)


# def Jy2AB(flux):
#     return flux2mag(mag, 8.90)


# def mJy2AB(flux):
#     return flux2mag(mag, 16.40)


# def uJy2AB(flux):
#     return flux2mag(mag, 23.90)


def AB2nMgy(mag):
    return mag2flux(mag, 22.50)


def AB2Jy(mag):
    return mag2flux(mag, 8.90)


def AB2mJy(mag):
    return mag2flux(mag, 16.40)


def AB2uJy(mag):
    return mag2flux(mag, 23.90)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-02-26
# @Filename: guide.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToTIC_v8,
    Gaia_DR2,
    Gaia_DR2_Clean,
    TIC_v8,
)

from .base import BaseCarton


Gaia = Gaia_DR2
Gaia_Clean = Gaia_DR2_Clean


class GuideCarton(BaseCarton):
    name = "guide"
    category = "guide"
    cadence = None

    def build_query(self, version_id, query_region=None):
        sample = (
            Catalog.select(Catalog.catalogid, Catalog.ra, Catalog.dec, Gaia.phot_g_mean_mag)
            .join(CatalogToTIC_v8)
            .join(TIC_v8)
            .join(Gaia)
            .join(Gaia_Clean)
            .where(
                (Gaia.phot_g_mean_mag > self.parameters["g_min"])
                & (Gaia.phot_g_mean_mag < self.parameters["g_max"])
            )
            .where(Catalog.version_id == version_id, CatalogToTIC_v8.version_id == version_id)
        )

        if query_region:
            sample = sample.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        sample = sample.cte("sample")

        # We should use q3c_join_pm and q3c_dist_pm here.
        subq = (
            Gaia.select(Gaia.source_id)
            .where(
                peewee.fn.q3c_join(
                    sample.c.ra, sample.c.dec, Gaia.ra, Gaia.dec, self.parameters["min_separation"]
                )
            )
            .order_by(peewee.fn.q3c_dist(sample.c.ra, sample.c.dec, Gaia.ra, Gaia.dec).asc())
            .limit(1)
            .offset(1)
        )

        decollided = (
            sample.select(sample.c.catalogid)
            .join(subq, peewee.JOIN.LEFT_LATERAL, on=peewee.SQL("true"))
            .where(subq.c.source_id.is_null())
            .with_cte(sample)
        )

        return decollided



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-06-17
# @Filename: bhm_spiders_clusters.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# isort: skip_file

import peewee
from peewee import JOIN
from peewee import fn


from target_selection.cartons.base import BaseCarton
from target_selection.mag_flux import AB2nMgy  # , AB2Jy

# general catalogdb imports
from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    EROSITASupersetv1Clusters,
)

# imports of existing spectro catalogues
from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogFromSDSS_DR19p_Speclite,
    SDSS_DR19p_Speclite,
)

# additional imports required by bhm_spiders_clusters_lsdr10
from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogToLegacy_Survey_DR10,
    Legacy_Survey_DR10,
)

# # additional imports required by bhm_spiders_clusters_ps1dr2
# from sdssdb.peewee.sdss5db.catalogdb import (
#     Panstarrs1,
#     CatalogToPanstarrs1,    # only exists after v0.5 cross-match
# )

# DEBUG STUFF TO USE TEMP TABLE
# CatalogToSDSS_DR19p_Speclite._meta.table_name = 'temp_catalog_to_sdss_dr19p_speclite'
# CatalogToSDSS_DR19p_Speclite._meta._schema = 'sandbox'

# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms

# ############################################
# ############################################
# ############################################
# ############################################
# # This file provides the following BHM cartons in v1.0:
#
#   bhm_spiders_clusters_lsdr10
#   bhm_spiders_clusters_lsdr10_d3
#
# ############################################
# ############################################
# ############################################
# ############################################

# Notes on how many targets to expect:
# => SELECT ero_version,xmatch_method,xmatch_version,opt_cat,count(*)
#      FROM erosita_superset_v1_clusters GROUP BY ero_version,xmatch_method,xmatch_version,opt_cat;
#
#  ero_version | xmatch_method | xmatch_version  | opt_cat |  count
# -------------+---------------+-----------------+---------+---------
#  eRASS4_020  | eromapper     | lsdr10a_grz_z_v | lsdr10  | 2330083

#
#
# END PREAMBLE
# ##################################################################################


class BhmSpidersClustersLsdr10Carton(BaseCarton):
    name = "bhm_spiders_clusters_lsdr10"
    category = "science"
    mapper = "BHM"
    program = "bhm_spiders"
    tile = False
    instrument = "BOSS"
    inertial = True
    can_offset = True
    only_faintest_cadence = False

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        ls = Legacy_Survey_DR10.alias()
        c2ls = CatalogToLegacy_Survey_DR10.alias()

        x = EROSITASupersetv1Clusters.alias()
        # xx = EROSITASupersetv1Clusters.alias()
        # x = (
        #     xx
        #     .select(
        #         fn.rank().over(partition_by=[xx.ero_detuid],
        #                        order_by=[xx.xmatch_metric.desc()]).alias('x_rank'),
        #         xx.ero_detuid.alias('ero_detuid'),
        #         xx.ls_id.alias('ls_id'),
        #         xx.target_has_spec.alias('target_has_spec'),
        #     )
        #     .where(
        #         (xx.ero_version == self.parameters['ero_version']),
        #         (xx.xmatch_method == self.parameters['xmatch_method']),
        #         (xx.xmatch_version == self.parameters['xmatch_version']),
        #         (xx.opt_cat == self.parameters['opt_cat']),
        #         (xx.xmatch_metric > self.parameters['xmatch_metric_min']),
        #         (xx.ero_det_like > self.parameters['det_like_min']),
        #     )
        #     .alias('x')
        # )

        instrument = peewee.Value(self.instrument)
        inertial = peewee.Value(self.inertial).cast("bool")

        fibertotflux_i_max = AB2nMgy(self.parameters["fibertotmag_i_min"])
        fibertotflux_i_min = AB2nMgy(self.parameters["fibertotmag_i_max"])
        fibertotflux_z_max = AB2nMgy(self.parameters["fibertotmag_z_min"])
        fibertotflux_z_min = AB2nMgy(self.parameters["fibertotmag_z_max"])
        fibertotflux_i_max_for_core = AB2nMgy(self.parameters["fibertotmag_i_min_for_core"])
        fibertotflux_i_min_for_core = AB2nMgy(self.parameters["fibertotmag_i_max_for_core"])
        fibertotflux_z_max_for_core = AB2nMgy(self.parameters["fibertotmag_z_min_for_core"])
        fibertotflux_z_min_for_core = AB2nMgy(self.parameters["fibertotmag_z_max_for_core"])

        fibertotflux_r_min_for_cadence1 = AB2nMgy(self.parameters["fibertotmag_r_for_cadence1"])
        fibertotflux_z_min_for_cadence1 = AB2nMgy(self.parameters["fibertotmag_z_for_cadence1"])
        fibertotflux_r_min_for_cadence2 = AB2nMgy(self.parameters["fibertotmag_r_for_cadence2"])
        gaia_g_max_for_cadence1 = self.parameters["gaia_g_max_for_cadence1"]
        gaia_rp_max_for_cadence1 = self.parameters["gaia_rp_max_for_cadence1"]

        # flux30 = AB2nMgy(30.00)

        # #########################################################################
        # prepare the spectroscopy catalogues
        spec_sn_thresh = self.parameters["spec_sn_thresh"]
        spec_z_err_thresh = self.parameters["spec_z_err_thresh"]

        # SDSS DR19p
        # downslect only 'good' spectra
        c2s19 = CatalogFromSDSS_DR19p_Speclite.alias()
        ss19 = SDSS_DR19p_Speclite.alias()
        s19 = (
            ss19.select(
                ss19.pk.alias("s19_pk"),
            )
            .where(
                ss19.sn_median_all >= spec_sn_thresh,
                ss19.zwarning == 0,
                ss19.z_err <= spec_z_err_thresh,
                ss19.z_err > 0.0,
                ss19.specprimary > 0,
            )
            .alias("s19")
        )
        # #########################################################################

        # logic is written this backwards way so that a failure to meet ant core
        # criterion results in non-core status
        is_core = peewee.Case(
            None,
            (
                (
                    ~(
                        (
                            ls.fiberflux_r.between(
                                fibertotflux_i_min_for_core, fibertotflux_i_max_for_core
                            )
                        )
                        | (
                            ls.fiberflux_i.between(
                                fibertotflux_z_min_for_core, fibertotflux_z_max_for_core
                            )
                        )
                    ),
                    False,
                ),
            ),
            True,
        )

        # priority is determined by target rank within cluster
        # start with a priority floor value (per carton)
        # then increment if any conditions are met:

        # priority = peewee.Case(
        #     None,
        #     (
        #         (
        #             x.c.x_rank == 1,
        #             self.parameters['priority_floor_bcg']
        #         ),
        #         (
        #             x.c.x_rank > 1,
        #             self.parameters['priority_floor_member'] +
        #             fn.least(self.parameters['priority_levels'] - 2,
        #                      x.c.x_rank - 2)
        #         ),
        #     ),
        #     None)

        dpriority_non_core = peewee.Case(
            None, ((is_core, 0),), self.parameters["dpriority_non_core"]
        )

        priority_floor = peewee.Case(
            None,
            ((x.target_priority == 0, self.parameters["priority_floor_bcg"]),),
            self.parameters["priority_floor_member"],
        )

        priority = priority_floor + x.target_priority + dpriority_non_core

        value = peewee.Case(
            None,
            (
                (~is_core, 0),
                (x.target_priority == 0, self.parameters["value_bcg"]),
                (x.target_priority > 0, self.parameters["value_member"]),
            ),
            None,
        ).cast("float")

        # choose cadence based on fiber magnitude in r-band
        cadence1 = self.parameters["cadence1"]
        cadence2 = self.parameters["cadence2"]
        cadence3 = self.parameters["cadence3"]
        cadence4 = "unknown_cadence"  # catch failures
        cadence = peewee.Case(
            None,
            (
                (
                    (
                        (ls.fibertotflux_r > fibertotflux_r_min_for_cadence1)
                        | (ls.fibertotflux_z > fibertotflux_z_min_for_cadence1)
                        | (ls.gaia_phot_g_mean_mag.between(0.1, gaia_g_max_for_cadence1))
                        | (ls.gaia_phot_rp_mean_mag.between(0.1, gaia_rp_max_for_cadence1))
                    ),
                    cadence1,
                ),
                (ls.fibertotflux_r > fibertotflux_r_min_for_cadence2, cadence2),
                (ls.fibertotflux_r <= fibertotflux_r_min_for_cadence2, cadence3),
            ),
            cadence4,
        )

        # compute transformed SDSS mags for pointlike and extended sources uniformly
        # transform the legacysurvey grz into sdss psfmag griz

        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ if($3~/sdss_psfmag/){pe="p"} else if ($3~/sdss_fiber2mag/){pe="e"} else{pe="error"}; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  bhm_spiders_clusters_lsdr8/lsdr8_fibermag_to_sdss_fiber2mag_?_results.log   # noqa
        coeffs = {
            "g2_e": -0.897719,
            "g1_e": 2.298300,
            "g0_e": -1.019299,
            "i2_e": -0.950114,
            "i1_e": 0.981972,
            "i0_e": -0.261645,
            "r2_e": -0.201741,
            "r1_e": 0.697128,
            "r0_e": -0.120926,
            "z2_e": -1.424312,
            "z1_e": 2.415301,
            "z0_e": -0.677163,
        }

        nMgy_min = 1e-3  # equiv to AB=30
        # extended - start from ls8 fiberfluxes
        g0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_g))
        r0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_r))
        i0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_i))
        z0_e = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fiberflux_z))

        g0_t = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fibertotflux_g))
        r0_t = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fibertotflux_r))
        i0_t = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fibertotflux_i))
        z0_t = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.fibertotflux_z))

        g0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_g))
        r0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r))
        i0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_i))
        z0_p = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_z))

        g_r_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_g)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
        )
        r_z_e = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.fiberflux_r)
            / peewee.fn.greatest(nMgy_min, ls.fiberflux_z)
        )

        g_e = g0_e + coeffs["g0_e"] + coeffs["g1_e"] * g_r_e + coeffs["g2_e"] * g_r_e * g_r_e
        r_e = r0_e + coeffs["r0_e"] + coeffs["r1_e"] * g_r_e + coeffs["r2_e"] * g_r_e * g_r_e
        i_e = r0_e + coeffs["i0_e"] + coeffs["i1_e"] * r_z_e + coeffs["i2_e"] * r_z_e * r_z_e
        z_e = z0_e + coeffs["z0_e"] + coeffs["z1_e"] * r_z_e + coeffs["z2_e"] * r_z_e * r_z_e

        # validity checks
        valid = g0_e.between(0.1, 29.9) & r0_e.between(0.1, 29.9) & z0_e.between(0.1, 29.9)

        opt_prov = peewee.Case(None, ((valid, "sdss_fiber2mag_from_lsdr10"),), "undefined")
        magnitude_g = peewee.Case(None, ((valid, g_e),), "NaN")
        magnitude_r = peewee.Case(None, ((valid, r_e),), "NaN")
        magnitude_i = peewee.Case(None, ((valid, i_e),), "NaN")
        magnitude_z = peewee.Case(None, ((valid, z_e),), "NaN")
        magnitude_gaia_g = peewee.Case(
            None, ((ls.gaia_phot_g_mean_mag.between(0.1, 29.9), ls.gaia_phot_g_mean_mag),), "NaN"
        )
        magnitude_gaia_bp = peewee.Case(
            None, ((ls.gaia_phot_bp_mean_mag.between(0.1, 29.9), ls.gaia_phot_bp_mean_mag),), "NaN"
        )
        magnitude_gaia_rp = peewee.Case(
            None, ((ls.gaia_phot_rp_mean_mag.between(0.1, 29.9), ls.gaia_phot_rp_mean_mag),), "NaN"
        )

        spec_sn_thresh = self.parameters["spec_sn_thresh"]
        spec_z_err_thresh = self.parameters["spec_z_err_thresh"]

        query = (
            c.select(
                c.catalogid.alias("catalogid"),
                ls.ls_id.alias("ls_id"),  # extra
                x.pkey.alias("ero_pkey"),  # extra
                x.ero_detuid.cast("text").alias("ero_detuid"),  # extra
                s19.c.s19_pk.alias("sdss_dr19p_speclite_pk"),  # extra
                c.ra.alias("ra"),  # extra
                c.dec.alias("dec"),  # extra
                priority.alias("priority"),
                value.alias("value"),
                cadence.alias("cadence"),
                instrument.alias("instrument"),
                opt_prov.alias("optical_prov"),
                magnitude_g.alias("g"),
                magnitude_r.alias("r"),
                magnitude_i.alias("i"),
                magnitude_z.alias("z"),
                magnitude_gaia_g.alias("gaia_g"),
                magnitude_gaia_bp.alias("bp"),
                magnitude_gaia_rp.alias("rp"),
                inertial.alias("inertial"),
                g0_p.alias("ls10_mag_g"),  # extra
                r0_p.alias("ls10_mag_r"),  # extra
                i0_p.alias("ls10_mag_i"),  # extra
                z0_p.alias("ls10_mag_z"),  # extra
                g0_e.alias("ls10_fibermag_g"),  # extra
                r0_e.alias("ls10_fibermag_r"),  # extra
                i0_e.alias("ls10_fibermag_i"),  # extra
                z0_e.alias("ls10_fibermag_z"),  # extra
                g0_t.alias("ls10_fibertotmag_g"),  # extra
                r0_t.alias("ls10_fibertotmag_r"),  # extra
                i0_t.alias("ls10_fibertotmag_i"),  # extra
                z0_t.alias("ls10_fibertotmag_z"),  # extra
                is_core.alias("is_core"),  # extra
                x.target_priority.alias("orig_target_priority"),
                x.eromapper_lambda.alias("eromapper_lambda"),
                x.eromapper_z_lambda.alias("eromapper_z_lambda"),
                x.xmatch_flags.alias("xmatch_flags"),
                x.target_has_spec.alias("target_has_spec"),
            )
            .join(c2ls)
            .join(ls)
            .join(x, on=(ls.ls_id == x.ls_id))
            # start joining the spectroscopy
            .switch(c)
            .join(c2s19, JOIN.LEFT_OUTER)
            .join(s19, JOIN.LEFT_OUTER, on=(s19.c.s19_pk == c2s19.target_id))
            # finished joining the spectroscopy
            .where(
                c.version_id == version_id,
                c2ls.version_id == version_id,
                c2ls.best >> True,
                fn.coalesce(c2s19.version_id, version_id) == version_id,
            )
            .where(
                s19.c.s19_pk.is_null(True),
            )
            .where(
                (x.ero_version == self.parameters["ero_version"]),
                (x.xmatch_method == self.parameters["xmatch_method"]),
                (x.xmatch_version == self.parameters["xmatch_version"]),
                (x.opt_cat == self.parameters["opt_cat"]),
                (x.xmatch_metric > self.parameters["xmatch_metric_min"]),
                (x.ero_det_like > self.parameters["det_like_min"]),
            )
            .where(
                (
                    (ls.fibertotflux_i.between(fibertotflux_i_min, fibertotflux_i_max))
                    | (ls.fibertotflux_z.between(fibertotflux_z_min, fibertotflux_z_max))
                ),
                (x.target_has_spec == 0),
                # gaia safety checks to avoid bad ls photometry
                ~(ls.gaia_phot_g_mean_mag.between(0.1, self.parameters["gaia_g_mag_limit"])),
                ~(ls.gaia_phot_rp_mean_mag.between(0.1, self.parameters["gaia_rp_mag_limit"])),
            )
            .distinct([ls.ls_id])
        )

        if self.only_faintest_cadence:
            query = query.where(cadence == cadence3)

        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    c.ra, c.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


#
# END BhmSpidersClustersLsdr10Carton
# ##################################################################################


class BhmSpidersClustersLsdr10D3Carton(BhmSpidersClustersLsdr10Carton):
    name = "bhm_spiders_clusters_lsdr10_d3"
    only_faintest_cadence = True
